@@ -1136,13 +1136,13 @@ VOID StGWRemove(gwinfo)
   if (gwinfo->FreeMem != NULL) (*gwinfo->FreeMem)(w);
 
   if (XDeleteContext(dpy, w, EventContext) != 0)
-    xlfail("cound not delete event context");
+    xlfail("could not delete event context");
   if (XDeleteContext(dpy, w, ObjectContext) != 0)
-    xlfail("cound not delete object context");
+    xlfail("could not delete object context");
   if (XDeleteContext(dpy, gwinfo->panel, EventContext) != 0)
-    xlfail("cound not delete event context");
+    xlfail("could not delete event context");
   if (XDeleteContext(dpy, gwinfo->panel, ObjectContext) != 0)
-    xlfail("cound not delete object context");
+    xlfail("could not delete object context");
 
   if (gwinfo->gc != NULL) XFreeGC(dpy, gwinfo->gc);
   gwinfo->gc = NULL;
