@@ -45,7 +45,7 @@
 # endif
 #endif /* THINK_C */
 #ifdef __MWERKS__
-# ifdef CHECK_FOR_MC68881
+# if (defined(__MC68881__) && __MC68881__)
 #  ifndef _MC68881_
 #   define _MC68881_
 #  endif _MC68881_
@@ -412,7 +412,7 @@ VOID osinit(char *name)
 	  XLS_MAJOR_RELEASE, XLS_MINOR_RELEASE, XLS_SUBMINOR_RELEASE,
 	  XLS_RELEASE_STATUS);
   TtyPrint(buf);
-  sprintf(buf, "Copyright (c) 1989-1998, by Luke Tierney.\n\n"); 
+  sprintf(buf, "Copyright (c) 1989-1999, by Luke Tierney.\n\n"); 
   TtyPrint(buf);
 }
 
@@ -443,7 +443,7 @@ VOID xoserror(char *msg)
   Str255 pbuf;
   NotifyIfInBackground();
   C2PSTRING(msg, pbuf);
-  FakeAlert ((StringPtr) pbuf, "\p", "\p", "\p", 1, 1, 0, "\pOK", "\p", "\p");
+  FakeAlert (pbuf, "\p", "\p", "\p", 1, 1, 0, "\pOK", "\p", "\p");
 }
 
 FILEP osaopen(const char *name, const char *mode)
@@ -787,7 +787,7 @@ int CintoPstring(const char *cp, StringPtr pp, int ppsize, int cvtnl)
   pp[0] = n;
   memmove(pp + 1, cp, n);
   if (cvtnl)
-    for (; n > 0; n--, pp++)
+    for (pp++; n > 0; n--, pp++)
       if (*pp == '\n')
         *pp = RETURNCHAR;
   return overflow;
@@ -967,17 +967,17 @@ VOID fsetfileinfo (char *fname, OSType creator, OSType type)
   GetDir(&dir);
   C2PSTRING(fname, pbuf);
   if (hasAppleEvents) {
-	FSMakeFSSpec(dir.vRefNum, dir.parID, (StringPtr) buf, &fsspec);
+	FSMakeFSSpec(dir.vRefNum, dir.parID, pbuf, &fsspec);
 	FSpGetFInfo(&fsspec, &finfo);
   }
   else
-    GetFInfo((StringPtr) buf, dir.vRefNum, &finfo);
+    GetFInfo(pbuf, dir.vRefNum, &finfo);
   finfo.fdCreator = creator;
   finfo.fdType = type;
   if (hasAppleEvents)
     FSpSetFInfo(&fsspec, &finfo);
   else
-    SetFInfo((StringPtr) buf, dir.vRefNum, &finfo);
+    SetFInfo(pbuf, dir.vRefNum, &finfo);
 }
 #endif /* applec */
 
@@ -1377,7 +1377,7 @@ LVAL xlaunchapp(void)
 
     GetDir(&dir);
 	C2PSTRING(getstring(arg), pbuf);
-	if (FSMakeFSSpec(dir.vRefNum, dir.parID, (StringPtr) buf, &fSpec) != noErr)
+	if (FSMakeFSSpec(dir.vRefNum, dir.parID, pbuf, &fSpec) != noErr)
 	  return NIL;
   }
   else if (fixp(arg)) {
