@@ -16,12 +16,20 @@ extern LVAL s_text_length;
 
 /* static variables */
 static char szXLSDlgClass[] = "XLSDialog";
+static char szXLSMDIDlgClass[] = "XLSMDIDialog";
 static char szXLSMDlgClass[] = "XLSModalDialog";
 static BOOL InModalDialog = FALSE;
 static LVAL ModalItem = NIL;
+#ifdef WIN32
+static WNDPROC pDefBtnProc = NULL;
+static WNDPROC pDefEditTextProc = NULL;
+static WNDPROC pDefScrollProc = NULL;
+static WNDPROC pDefListProc = NULL;
+#endif /* WIN32 */
 
 /* function prototypes */
 LRESULT CALLBACK XLSDlgProc(HWND, UINT, WPARAM, LONG);
+LRESULT CALLBACK XLSMDIDlgProc(HWND, UINT, WPARAM, LONG);
 LRESULT CALLBACK XLSMDlgProc(HWND, UINT, WPARAM, LONG);
 static void InstallItem(HWND, LVAL);
 static void InstallButtonItem(HWND, LVAL);
@@ -31,6 +39,29 @@ static void InstallTextItem(HWND, LVAL);
 static void InstallScrollItem(HWND, LVAL);
 static void InstallListItem(HWND, LVAL);
 static void SetClusterValue(HWND, int);
+#ifdef WIN32
+static void DialogDoDefaultButton(LVAL dialog);
+static LRESULT CALLBACK CustomButtonProc(HWND hWnd,UINT message,WPARAM wPrm,LPARAM lPrm);
+static LRESULT CALLBACK CustomEditTextProc(HWND hWnd,UINT message,WPARAM wPrm,LPARAM lPrm);
+static LRESULT CALLBACK CustomScrollProc(HWND hWnd,UINT message,WPARAM wPrm,LPARAM lPrm);
+static LRESULT CALLBACK CustomListProc(HWND hWnd,UINT message,WPARAM wPrm,LPARAM lPrm);
+#endif /* WIN32 */
+
+#ifdef WIN32
+#  define DIALOG_FONT DEFAULT_GUI_FONT
+#  define DIALOG_BACKGROUND COLOR_MENU
+#  define AVE_CHAR_PAD 2
+#  define CHOICE_WIDTH_FUZZ 6
+#else
+#  define DIALOG_FONT SYSTEM_FIXED_FONT
+#  define DIALOG_BACKGROUND COLOR_WINDOW
+#  define AVE_CHAR_PAD 0
+#  define CHOICE_WIDTH_FUZZ 4
+#endif
+#define LIST_USE_LEADING FALSE  /**** inline this */
+#define REAL_LIST_ITEM_PAD 0 /**** drop LIST_ITEM_PAD */
+/**** need to rethink size calculations for non-fixed-width fonts */
+
 
 /***********************************************************************/
 /**                                                                   **/
@@ -90,9 +121,9 @@ static Point AvCharSize(BOOL leaded)
   TEXTMETRIC tm;
 
   hDC = GetDC(hWndFrame);
-  SelectObject(hDC, GetStockObject(SYSTEM_FIXED_FONT));
+  SelectObject(hDC, GetStockObject(DIALOG_FONT));
   GetTextMetrics(hDC, &tm);
-  pt.h = tm.tmAveCharWidth;
+  pt.h = tm.tmAveCharWidth + AVE_CHAR_PAD; /* to correct font width problems */
   pt.v = tm.tmHeight;
   if (leaded) pt.v += tm.tmExternalLeading;
   ReleaseDC(hWndFrame, hDC);
@@ -242,11 +273,24 @@ BOOL InitApplDialogs(HANDLE hInstance)
   wc.cbClsExtra = 0;
   wc.cbWndExtra = sizeof(LONG) + sizeof(LONG);
   wc.hInstance = hInstance;
-  wc.hIcon = NULL;
-  wc.hCursor = LoadCursor((HWND) NULL, IDC_ARROW);
-  wc.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1);
+  wc.hIcon = LoadIcon(hInstance, "WXLSIcon");
+  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
+  wc.hbrBackground = (HBRUSH) (DIALOG_BACKGROUND + 1);
   wc.lpszMenuName = NULL;
-  wc. lpszClassName = szXLSDlgClass;
+  wc.lpszClassName = szXLSDlgClass;
+  if (! RegisterClass(&wc)) return(FALSE);
+
+  /* class structure for modeless MDI dialog windows */
+  wc.style = CS_HREDRAW | CS_VREDRAW;
+  wc.lpfnWndProc = (WNDPROC) XLSMDIDlgProc;
+  wc.cbClsExtra = 0;
+  wc.cbWndExtra = sizeof(LONG) + sizeof(LONG);
+  wc.hInstance = hInstance;
+  wc.hIcon = LoadIcon(hInstance, "WXLSIcon");
+  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
+  wc.hbrBackground = (HBRUSH) (DIALOG_BACKGROUND + 1);
+  wc.lpszMenuName = NULL;
+  wc.lpszClassName = szXLSMDIDlgClass;
   if (! RegisterClass(&wc)) return(FALSE);
 
   /* class structure for modal dialog windows */
@@ -256,16 +300,32 @@ BOOL InitApplDialogs(HANDLE hInstance)
   wc.cbWndExtra = sizeof(LONG) + sizeof(LONG);
   wc.hInstance = hInstance;
   wc.hIcon = NULL;
-  wc.hCursor = LoadCursor((HWND) NULL, IDC_ARROW);
-  wc.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1);
+  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
+  wc.hbrBackground = (HBRUSH) (DIALOG_BACKGROUND + 1);
   wc.lpszMenuName = NULL;
-  wc. lpszClassName = szXLSMDlgClass;
+  wc.lpszClassName = szXLSMDlgClass;
   if (! RegisterClass(&wc)) return(FALSE);
+#ifdef WIN32
+  //
+  // these procs are used to pass the ENTER key on to the parent window to
+  // mimic default button modal dialog behavior
+  //
+  GetClassInfo(hInstance, "button", &wc);
+  pDefBtnProc = wc.lpfnWndProc;
+
+  GetClassInfo(hInstance, "edit", &wc);
+  pDefEditTextProc = wc.lpfnWndProc;
+
+  GetClassInfo(hInstance, "scrollbar", &wc);
+  pDefScrollProc = wc.lpfnWndProc;
+
+  GetClassInfo(hInstance, "listbox", &wc);
+  pDefListProc = wc.lpfnWndProc;
+#endif /* WIN32 */
   return(TRUE);
 }
 
-InitInstDialogs(HANDLE hInstance)
-#pragma argsused
+int InitInstDialogs(HANDLE hInstance)
 {
   return(TRUE);
 }
@@ -274,7 +334,8 @@ static LRESULT ProcessDialogEvent(HWND hWnd,
   		                  UINT message,
 			          WPARAM wPrm,
 			          LONG lPrm,
-                                  BOOL modeless)
+                                  BOOL modeless,
+				  BOOL mdidialog)
 {
 
   LVAL dialog;
@@ -282,8 +343,17 @@ static LRESULT ProcessDialogEvent(HWND hWnd,
   int val, wid;
 
   switch(message) {
+#ifdef WIN32
+    /* Set the background color for WIN32 */
+  case WM_CTLCOLORDLG:
+  case WM_CTLCOLORSTATIC:
+  case WM_CTLCOLORBTN:
+    val = SetBkColor((HDC) wPrm, GetSysColor(DIALOG_BACKGROUND));
+    SelectObject((HDC) wPrm, GetStockObject(DIALOG_FONT));
+    return ((LRESULT) GetSysColorBrush(DIALOG_BACKGROUND));
+#endif
   case WM_CREATE:
-    if (modeless) {
+    if (mdidialog) {
       LPMDICREATESTRUCT mdic;
       mdic = (LPMDICREATESTRUCT) ((LPCREATESTRUCT) lPrm)->lpCreateParams;
       dialog = (LVAL) mdic->lParam;
@@ -299,9 +369,15 @@ static LRESULT ProcessDialogEvent(HWND hWnd,
     switch (GET_WM_COMMAND_ID(wPrm, lPrm)) {
     case IDC_SHOWWINDOW:
       if (modeless) {
-        if (IsIconic(hWnd))
-	  MDIRestoreWindow(hWndClient, hWnd);
-        MDIActivateWindow(hWndClient, hWnd);
+	if (mdidialog) {
+	  if (IsIconic(hWnd))
+	    MDIRestoreWindow(hWndClient, hWnd);
+	  MDIActivateWindow(hWndClient, hWnd);
+	}
+	else {
+	  if (IsIconic(hWnd))
+	    ShowWindow(hWnd,SW_RESTORE);
+	}
         return(0);
       }
       break;
@@ -313,7 +389,7 @@ static LRESULT ProcessDialogEvent(HWND hWnd,
       }
       break;
     case IDC_DESTROY:
-      if (modeless)
+      if (mdidialog)
         MDIDestroyWindow(hWndClient, hWnd);
       else
         DestroyWindow(hWnd);
@@ -400,6 +476,19 @@ static LRESULT ProcessDialogEvent(HWND hWnd,
       return(0);
     }
     break;
+#ifdef WIN32
+  case WM_CHAR:
+    switch (wPrm) {
+    case VK_RETURN:
+      // do the default button thing
+      DialogDoDefaultButton(GETDIALOGOBJECT(hWnd));
+      break;
+    case VK_ESCAPE:
+      DestroyWindow(hWnd);
+      break;
+    }
+    break;
+#endif /* WIN32 */
   case WM_QUERYENDSESSION:
   case WM_CLOSE:
     if (! InModalDialog)
@@ -407,28 +496,27 @@ static LRESULT ProcessDialogEvent(HWND hWnd,
     return(0);
   case WM_DESTROY:
     FreeDialogData(GETDIALOGDATA(hWnd));
-    if (modeless) return 0;
+    if (mdidialog) return 0;
   }
-  if (modeless)
+  if (mdidialog)
     return(DefMDIChildProc(hWnd, message, wPrm, lPrm));
   else
     return(DefWindowProc(hWnd, message, wPrm, lPrm));
 }
 
-LRESULT CALLBACK XLSDlgProc(HWND hWnd,
-		            UINT message,
-		            WPARAM wPrm,
-		            LONG lPrm)
+LRESULT CALLBACK XLSDlgProc(HWND hWnd, UINT message, WPARAM wPrm, LONG lPrm)
 {
-  return ProcessDialogEvent(hWnd, message, wPrm, lPrm, TRUE);
+  return ProcessDialogEvent(hWnd, message, wPrm, lPrm, TRUE, FALSE);
 }
 
-LRESULT CALLBACK XLSMDlgProc(HWND hWnd,
-		             UINT message,
-			     WPARAM wPrm,
-			     LONG lPrm)
+LRESULT CALLBACK XLSMDIDlgProc(HWND hWnd, UINT message, WPARAM wPrm, LONG lPrm)
 {
-  return ProcessDialogEvent(hWnd, message, wPrm, lPrm, FALSE);
+  return ProcessDialogEvent(hWnd, message, wPrm, lPrm, TRUE, TRUE);
+}
+
+LRESULT CALLBACK XLSMDlgProc(HWND hWnd, UINT message, WPARAM wPrm, LONG lPrm)
+{
+  return ProcessDialogEvent(hWnd, message, wPrm, lPrm, FALSE, FALSE);
 }
 
 void MSWResetDialogs(void)
@@ -465,7 +553,10 @@ void DialogAllocate(LVAL dialog)
   if (size.h < MIN_DLG_WIDTH) size.h = MIN_DLG_WIDTH;
   if (size.v < MIN_DLG_HEIGHT) size.v = MIN_DLG_HEIGHT;
   if (modeless) { // ### fix if manage to get proper frame
-    MDICREATESTRUCT mdicreate;
+    /**** This is a quick hack.  It would probably be better to use a slot
+	  to set the MDI/non-MDI option */
+    LVAL topsym = xlenter("SYSTEM::*TOP-LEVEL-DIALOGS*");
+    int mdidialog = (! boundp(topsym)) || null(getvalue(topsym));
 
     size.h += 2 * GetSystemMetrics(SM_CXBORDER);
     size.v += GetSystemMetrics(SM_CYBORDER);
@@ -475,33 +566,49 @@ void DialogAllocate(LVAL dialog)
     flags = WS_POPUP | WS_DLGFRAME | WS_VISIBLE | WS_CAPTION;
     if (goAway) flags |= WS_SYSMENU;
 
-    mdicreate.szClass = szXLSDlgClass;
-    mdicreate.szTitle = title;
-    mdicreate.hOwner = hInst;
-    mdicreate.x = loc.h;
-    mdicreate.y = loc.v;
-    mdicreate.cx = size.h;
-    mdicreate.cy = size.v;
-    mdicreate.style = WS_MINIMIZE;
-    mdicreate.lParam = (LONG) dialog;
-    theDialog = MDICreateWindow(hWndClient, &mdicreate);
+    if (mdidialog) {
+      MDICREATESTRUCT mdicreate;
 
-    SetWindowStyle(theDialog, GetWindowStyle(theDialog)
-                              & ~WS_MAXIMIZEBOX
-                              & ~WS_MINIMIZEBOX
-                              /*& ~WS_THICKFRAME*/);
-    MDIRestoreWindow(hWndClient, theDialog);
+      mdicreate.szClass = szXLSMDIDlgClass;
+      mdicreate.szTitle = title;
+      mdicreate.hOwner = hInst;
+      mdicreate.x = loc.h;
+      mdicreate.y = loc.v;
+      mdicreate.cx = size.h;
+      mdicreate.cy = size.v;
+      mdicreate.style = WS_MINIMIZE;
+      mdicreate.lParam = (LONG) dialog;
+
+      theDialog = MDICreateWindow(hWndClient, &mdicreate);
+      SetWindowStyle(theDialog, GetWindowStyle(theDialog)
+		     & ~WS_MAXIMIZEBOX & ~WS_MINIMIZEBOX);
+      MDIRestoreWindow(hWndClient, theDialog);
+    }
+    else {
+      theDialog = CreateWindowEx(WS_EX_DLGMODALFRAME, szXLSDlgClass,
+				 title, flags,
+				 loc.h, loc.v, size.h, size.v,
+				 NULL, NULL, hInst, (LPVOID) dialog);
+    }
   }
   else {
     size.h += 2 * GetSystemMetrics(SM_CXDLGFRAME);
+#ifdef WIN32
+    size.v += GetSystemMetrics(SM_CYBORDER);
+    size.v += GetSystemMetrics(SM_CYCAPTION);
+    flags = WS_POPUP | WS_VISIBLE |WS_CAPTION;      //FC 7/20/99     
+    theDialog = CreateWindowEx(WS_EX_DLGMODALFRAME, szXLSMDlgClass,
+			       title, flags,
+			       loc.h, loc.v, size.h, size.v,
+			       NULL, NULL, hInst, (LPVOID) dialog);
+#else
     size.v += 2 * GetSystemMetrics(SM_CYDLGFRAME);
-
     flags = WS_POPUP | WS_DLGFRAME | WS_VISIBLE;
-
     theDialog = CreateWindow(szXLSMDlgClass,
 			     title, flags,
 			     loc.h, loc.v, size.h, size.v,
 			     hWndFrame, NULL, hInst, (LPVOID) dialog);
+#endif
   }
 
   check_alloc(theDialog);
@@ -562,7 +669,9 @@ static void InstallBtnItem(HWND theDialog,
   data[itemIndex].itemHandle = theItem;
   data[itemIndex].object = item;
   if (di) *di = data[itemIndex];
-
+#ifdef WIN32
+  SetWindowLong(theItem, GWL_WNDPROC, (LPARAM)CustomButtonProc);
+#endif /* WIN32 */
   GlobalUnlock(hd);
 }
 
@@ -571,6 +680,48 @@ static void InstallButtonItem(HWND theDialog, LVAL item)
   InstallBtnItem(theDialog, item, BS_PUSHBUTTON, BUTTON_ITEM, NULL);
 }
 
+#ifdef WIN32
+static LRESULT CALLBACK CustomButtonProc(HWND hWnd, UINT message,
+					 WPARAM wPrm, LPARAM lPrm)
+{
+  if ((message == WM_CHAR) && (wPrm == VK_RETURN)) {
+    SendMessage(GetParent(hWnd), message, wPrm, lPrm);
+    return 0;
+  }
+  return CallWindowProc(pDefBtnProc, hWnd, message, wPrm, lPrm);
+}
+
+static LRESULT CALLBACK CustomScrollProc(HWND hWnd, UINT message,
+					 WPARAM wPrm, LPARAM lPrm)
+{
+  if ((message == WM_CHAR) && (wPrm == VK_RETURN)) {
+    SendMessage(GetParent(hWnd), message, wPrm, lPrm);
+    return 0;
+  }
+  return CallWindowProc(pDefScrollProc, hWnd, message, wPrm, lPrm);
+}
+
+static LRESULT CALLBACK CustomEditTextProc(HWND hWnd, UINT message,
+					   WPARAM wPrm, LPARAM lPrm)
+{
+  if ((message == WM_CHAR) && (wPrm == VK_RETURN)) {
+    SendMessage(GetParent(hWnd), message, wPrm, lPrm);
+    return 0;
+  }
+  return CallWindowProc(pDefEditTextProc, hWnd, message, wPrm, lPrm);
+}
+
+static LRESULT CALLBACK CustomListProc(HWND hWnd, UINT message,
+				       WPARAM wPrm, LPARAM lPrm)
+{
+  if ((message == WM_CHAR) && (wPrm == VK_RETURN)) {
+    SendMessage(GetParent(hWnd), message, wPrm, lPrm);
+    return 0;
+  }
+  return CallWindowProc(pDefListProc, hWnd, message, wPrm, lPrm);
+}
+#endif /* WIN32 */
+
 void DialogButtonGetDefaultSize(LVAL item, int *pwidth, int *pheight)
 {
   LVAL text;
@@ -672,6 +823,9 @@ static void InstallChoiceItem(HWND theDialog, LVAL item)
     data[itemIndex].object = item;
     data[itemIndex].clusterLeader = clusterLeader;
     data[itemIndex].clusterSize = clusterSize;
+#ifdef WIN32
+    SetWindowLong(theItem, GWL_WNDPROC, (LPARAM)CustomButtonProc);
+#endif /* WIN32 */
   }
   temp = slot_value(item, s_value);
   initial = (fixp(temp)) ? (int) getfixnum(temp) : 0;
@@ -693,7 +847,7 @@ void DialogChoiceGetDefaultSize(LVAL item, int *pwidth, int *pheight)
     if (n < k) n = k;
     h += (3 * csz.v) / 2;
   }
-  if (pwidth != NULL) *pwidth = (n + 4) * csz.h;
+  if (pwidth != NULL) *pwidth = (n + CHOICE_WIDTH_FUZZ) * csz.h;
   if (pheight != NULL) *pheight = h;
 }
 
@@ -782,7 +936,11 @@ static void InstallTextItem(HWND theDialog, LVAL item)
 			   hInst, NULL);
   }
   check_alloc(theItem);
-
+#ifdef WIN32  
+  if (editable) {
+   SetWindowLong(theItem, GWL_WNDPROC, (LPARAM)CustomEditTextProc);
+  }
+#endif /* WIN32 */
   data = GETITEMDATA(dialogData);
   data[itemIndex].type = TEXT_ITEM;
   data[itemIndex].itemNumber = itemIndex;
@@ -878,10 +1036,12 @@ static void InstallScrollItem(HWND theDialog, LVAL item)
   data[itemIndex].itemHandle = theItem;
   data[itemIndex].object = item;
   GlobalUnlock(hd);
+#ifdef WIN32
+  SetWindowLong(theItem, GWL_WNDPROC, (LPARAM)CustomScrollProc);
+#endif /* WIN32 */
 }
 
 void DialogScrollGetDefaultSize(LVAL item, int *pwidth, int *pheight)
-#pragma argsused pwidth pheight
 {
   int h, w;
 
@@ -979,8 +1139,8 @@ static void InstallListItem(HWND theDialog, LVAL item)
   DWORD flags;
   char *s;
 
-  csz = AvCharSize(TRUE);
-  csz.v += LIST_ITEM_PAD;
+  csz = AvCharSize(LIST_USE_LEADING);
+  csz.v += REAL_LIST_ITEM_PAD;
   loc = ListToPoint(slot_value(item, s_location));
   size = ListToPoint(slot_value(item, s_size));
 
@@ -995,7 +1155,7 @@ static void InstallListItem(HWND theDialog, LVAL item)
 
   hscroll = (columns < m) ? TRUE : FALSE;
   vscroll = (n * csz.v > size.v - ((hscroll) ? GetSystemMetrics(SM_CYHSCROLL) : 0)) ? TRUE : FALSE;
-  flags = WS_CHILD | WS_VISIBLE | WS_BORDER | LBS_NOTIFY;
+  flags = WS_CHILD | WS_VISIBLE | WS_BORDER | LBS_NOTIFY | LBS_NOINTEGRALHEIGHT;
   if (hscroll) flags |= WS_HSCROLL;
   if (vscroll) flags |= WS_VSCROLL;
   if (m > 1) flags |= LBS_MULTICOLUMN;
@@ -1012,6 +1172,7 @@ static void InstallListItem(HWND theDialog, LVAL item)
                          (HMENU) (itemIndex + ITEM_INDEX_BASE),
                          hInst, NULL);
   check_alloc(theItem);
+  SendMessage(theItem,WM_SETFONT,(WPARAM) GetStockObject(DIALOG_FONT),0);
 
   ListBox_SetColumnWidth(theItem, LIST_COL_CHARS * csz.h);
   if (darrayp(listData))
@@ -1037,6 +1198,9 @@ static void InstallListItem(HWND theDialog, LVAL item)
   data[itemIndex].object = item;
   GlobalUnlock(hd);
   xlpop();
+#ifdef WIN32
+  SetWindowLong(theItem, GWL_WNDPROC, (LPARAM)CustomListProc);
+#endif /* WIN32 */
 }
 
 void DialogListGetDefaultSize(LVAL item, int *pwidth, int *pheight)
@@ -1045,14 +1209,14 @@ void DialogListGetDefaultSize(LVAL item, int *pwidth, int *pheight)
   LVAL data = slot_value(item, s_list_data);
   Point csz, sz;
 
-  csz = AvCharSize(TRUE);
+  csz = AvCharSize(LIST_USE_LEADING);
   sz.h = LIST_COL_CHARS * (short) getfixnum(columns);
   if (seqp(data))
     sz.v = (short) seqlen(data);
   else if (matrixp(data))
     sz.v = (int) numrows(data);
 
-  csz.v += LIST_ITEM_PAD;
+  csz.v += REAL_LIST_ITEM_PAD;
   sz.h *= csz.h;
   sz.v *= csz.v;
   if (sz.v > csz.v * MAX_LIST_ROWS) {
@@ -1189,6 +1353,38 @@ void DialogSetDefaultButton(LVAL dialog, LVAL item)
   }
 }
 
+#ifdef WIN32
+static void DialogDoDefaultButton(LVAL dialog)
+{
+  HANDLE hd;
+  DialogData *dd;
+  DialogItemData di;
+  HWND theDialog;
+  LVAL item;
+
+  item = slot_value(dialog, s_default_button);
+    
+  if (item != NIL && ! button_item_p(item))
+    xlerror("not a button item", item);
+
+  theDialog = GETDIALOGADDRESS(dialog);
+  if (theDialog != NULL) {
+    hd = GETDIALOGDATA(theDialog);
+    dd = (DialogData *) GlobalLock(hd);
+    check_lock(dd);
+
+    if (FindItemData(theDialog, ITEM_INDEX_BASE + dd->dflt, &di))
+      dd->dflt = -1;
+    if (item != NIL && FindItemObjectData(item, &di)) {
+      HWND hwndDefault = di.itemHandle;
+      if (IsWindow(hwndDefault))
+	PostMessage(hwndDefault, BM_CLICK, 0, 0);
+    }
+    GlobalUnlock(hd);
+  }
+}
+#endif /* WIN32 */
+
 #ifdef DODO
 Add keyboard support:
 subclass all items to handle tabs
