@@ -24,7 +24,7 @@ static BOOL MouseCaptured = FALSE, MouseMotionOnly;
 static HWND MouseWnd = 0;
 static void (*MouseAction)(HWND, int, int);
 static TEXTMETRIC tm;
-static HFONT hGraphFont;
+static HFONT hGraphFont, hGraphFontUp;
 
 /* buffer variables */
 static HBITMAP monoBuffBits = 0, colorBuffBits = 0;
@@ -44,6 +44,7 @@ static void init_msw_text(void);
 static void cleanup_msw_text(void);
 static void init_msw_symbols(void);
 static void cleanup_msw_symbols(void);
+static void StGWPrint(StGWWinInfo *gwinfo);
 
 /**************************************************************************/
 /**                                                                      **/
@@ -80,7 +81,7 @@ void StGWInitWinInfo(LVAL object)
   gwinfo->cursor = 0;
   gwinfo->RefCon = 0L;
   gwinfo->drawPen.lopnStyle = PS_SOLID;
-  gwinfo->drawPen.lopnWidth.x = 0;
+  gwinfo->drawPen.lopnWidth.x = 1;
   gwinfo->drawPen.lopnColor = 0;
   gwinfo->rect_offset = 0;
 }
@@ -225,8 +226,6 @@ HWND IViewWindowNew(LVAL object, int is_GW)
   wind = MDICreateWindow(hWndClient, &mdicreate);
   if (! wind) xlfail("allocation Failed");
 
-  SetWindowStyle(wind, GetWindowStyle(wind) & ~WS_MAXIMIZEBOX);
-
   gwinfo = (StGWWinInfo *) StGWObWinInfo(object);
   gwinfo->window = wind;
   SetScrollRange(wind, SB_VERT, 0, 0, TRUE);
@@ -274,7 +273,7 @@ void CaptureMouse(HWND hWnd)
 long CALLBACK IVWinProc(HWND hWnd, UINT message, WPARAM wParam, LONG lParam)
 {
   StGWWinInfo *gwinfo;
-  static activating = FALSE;
+  static int activating = FALSE;
 
   switch (message) {
   case WM_CREATE:
@@ -295,6 +294,11 @@ long CALLBACK IVWinProc(HWND hWnd, UINT message, WPARAM wParam, LONG lParam)
       gwinfo = GETGWINFO(hWnd);
       StGWCopyToClip(gwinfo);
       return(0);
+    case IDM_PRINT:
+      gwinfo  = GETGWINFO(hWnd);
+      StGWPrint(gwinfo);
+    break;
+      break;
     }
     break;
   case WM_SIZE:
@@ -312,11 +316,6 @@ long CALLBACK IVWinProc(HWND hWnd, UINT message, WPARAM wParam, LONG lParam)
 	StGWSetScroll(gwinfo, gwinfo->view_h, gwinfo->view_v, TRUE);
       StGWObResize(GETWINOBJECT(hWnd));
     }
-    if (! IsIconic(hWnd)) { // to avoid problems with maximizing
-      DefMDIChildProc(hWnd, message, wParam, lParam);
-      MDIRestoreWindow(hWndClient, hWnd);
-      return(0);
-    }
     // ### fix slot value
     break;
   case WM_MOVE:
@@ -335,6 +334,22 @@ long CALLBACK IVWinProc(HWND hWnd, UINT message, WPARAM wParam, LONG lParam)
       case SB_LINEDOWN:      val += incs[0];       break;
       case SB_PAGEUP:        val -= incs[1];       break;
       case SB_PAGEDOWN:      val += incs[1];       break;
+      case SB_TOP:
+	{
+	  int nMin,nMax;
+	  int scroll = (message == WM_HSCROLL) ? SB_HORZ : SB_VERT;
+	  GetScrollRange(hWnd, scroll, &nMin, &nMax);
+	  val = nMin;
+	  break;
+	}
+      case SB_BOTTOM:
+	{
+	  int nMin,nMax;
+	  int scroll = (message == WM_HSCROLL) ? SB_HORZ : SB_VERT;
+	  GetScrollRange(hWnd, scroll, &nMin, &nMax);
+	  val = nMax;
+	  break;
+	}
       case SB_THUMBPOSITION: val = GET_WM_VSCROLL_POS(wParam,lParam); break;
       }
       hval = (message == WM_HSCROLL) ? val: gwinfo->view_h;
@@ -348,6 +363,27 @@ long CALLBACK IVWinProc(HWND hWnd, UINT message, WPARAM wParam, LONG lParam)
 		GetKeyState(VK_SHIFT),
 		GetKeyState(VK_CONTROL));
     return(0);
+#ifdef DODO
+    /**** These don't make any sense.  Something like sending a symbol
+      or a logical key code to the application is needed */
+    case WM_KEYDOWN:
+     switch (wParam) {
+     case VK_UP:
+     case VK_DOWN:
+     case VK_LEFT:
+     case VK_RIGHT:
+     case VK_END:
+     case VK_HOME:
+     case VK_PRIOR:
+     case VK_NEXT:
+       StGWObDoKey(GETWINOBJECT(hWnd),
+		  (unsigned char) wParam,
+		   GetKeyState(VK_SHIFT),
+		   GetKeyState(VK_CONTROL));
+       break;
+     }
+     return(0);
+#endif
   case WM_MOUSEMOVE:
     gwinfo = GETGWINFO(hWnd);
     SetCursor(get_cursor(gwinfo->cursor));
@@ -549,13 +585,15 @@ void StGWSetIdleOn(StGWWinInfo *gwinfo, int on)
 //**** not sure this is needed -- it probably duplicates stuff
 static void ResetDCClipRect(HDC hDC, StGWWinInfo *gwinfo)
 {
-  HRGN rgn;
-  rgn = CreateRectRgn(gwinfo->clip_rect.left - gwinfo->view_h,
-		      gwinfo->clip_rect.top - gwinfo->view_v,
-		      gwinfo->clip_rect.right - gwinfo->view_h,
-		      gwinfo->clip_rect.bottom) - gwinfo->view_v;
-  SelectClipRgn(hDC, rgn);
-  DeleteObject(rgn);
+  if (gwinfo->clipped) {
+    HRGN rgn;
+    rgn = CreateRectRgn(gwinfo->clip_rect.left - gwinfo->view_h,
+                        gwinfo->clip_rect.top - gwinfo->view_v,
+		        gwinfo->clip_rect.right - gwinfo->view_h,
+		        gwinfo->clip_rect.bottom) - gwinfo->view_v;
+    SelectClipRgn(hDC, rgn);
+    DeleteObject(rgn);
+  }
 }
 
 int StGWCanvasWidth(StGWWinInfo *gwinfo) { return (gwinfo ? gwinfo->canvasWidth : 0); }
@@ -751,7 +789,7 @@ void StGWSetCursor(StGWWinInfo *gwinfo, int cursor)
   if (gwinfo) gwinfo->cursor = cursor;
 }
 
-StGWCursor(StGWWinInfo *gwinfo)
+int StGWCursor(StGWWinInfo *gwinfo)
 {
   return(gwinfo ? gwinfo->cursor : 0);
 }
@@ -823,7 +861,7 @@ static void free_bitmap_bits(char *bits)
   if (bits) StFree(bits);
 }
 
-StGWMakeCursor(int n, char *image, char *mask, int h, int v, long refcon)
+int StGWMakeCursor(int n, char *image, char *mask, int h, int v, long refcon)
 {
   int index, cwidth, cheight, i;
   char *imageBits, *maskBits;
@@ -851,7 +889,7 @@ StGWMakeCursor(int n, char *image, char *mask, int h, int v, long refcon)
   return(index);
 }
 
-StGWMakeResCursor(char *name, int num, long refcon)
+int StGWMakeResCursor(char *name, int num, long refcon)
 {
   int index;
   char *cname = NULL;
@@ -1066,7 +1104,7 @@ void StGWSetLineWidth(StGWWinInfo *gwinfo, int width)
   if (gwinfo) {
     gwinfo->lineWidth = (width > 0) ? width : 1;
     gwinfo->rect_offset = width / 2;
-    gwinfo->drawPen.lopnWidth.x = (width > 1) ? width : 0;
+    gwinfo->drawPen.lopnWidth.x = (width > 1) ? width : 1;
   }
 }
 
@@ -1112,7 +1150,7 @@ void StGWSetClipRect(StGWWinInfo *gwinfo, int clipping,
     right -= gwinfo->view_h;
     bottom -= gwinfo->view_v;
     hDC = GetDC((HWND) gwinfo->window);
-    rgn = CreateRectRgn(left, top, right, bottom); //**** error check
+    rgn = CreateRectRgn(left, top, right, bottom); /***** error check */
     SelectClipRgn(hDC, rgn);
     ReleaseDC((HWND) gwinfo->window, hDC);
     if (currentDC) SelectClipRgn(currentDC, rgn);
@@ -1147,11 +1185,37 @@ void init_msw_buffer(void)
   }
 }
 
+static void init_dc(StGWWinInfo *gwinfo, HDC bufferDC, int buffer)
+{
+  HBRUSH hbr;
+  RECT r;
+  SetBkMode(bufferDC, TRANSPARENT);
+  SetBkColor(bufferDC, get_color(gwinfo->backColor));
+  SetTextColor(bufferDC, get_color(gwinfo->drawColor));
+  SelectObject(bufferDC, hGraphFont);
+  SetROP2(bufferDC, gwinfo->drawMode ? R2_NOT : R2_COPYPEN);
+  SetPolyFillMode(bufferDC, WINDING);
+  if (buffer)
+    SetViewportOrgEx(bufferDC, -gwinfo->view_h, -gwinfo->view_v, NULL);
+  ResetDCClipRect(bufferDC, gwinfo);
+  if (buffer) {
+    GetClientRect((HWND) gwinfo->window, &r);
+    r.left += gwinfo->view_h; r.top += gwinfo->view_v;
+    r.right += gwinfo->view_h; r.bottom += gwinfo->view_v;
+  }
+  else {
+    r.left = r.top = 0;
+    r.right = gwinfo->canvasWidth;
+    r.bottom = gwinfo->canvasHeight;
+  }
+  hbr = GET_ERASE_BRUSH(gwinfo);
+  FillRect(bufferDC, &r, hbr);
+  RELEASE_BRUSH(hbr);
+}
+
 void StGWStartBuffering(StGWWinInfo *gwinfo)
 {
   HDC hDC;
-  HBRUSH hbr;
-  RECT r;
 
   if (gwinfo && gwinfo->window) {
     if (! buffering) {
@@ -1161,20 +1225,7 @@ void StGWStartBuffering(StGWWinInfo *gwinfo)
       bufferDC = CreateCompatibleDC(hDC);
       ReleaseDC((HWND) gwinfo->window, hDC);
       SelectObject(bufferDC, (gwinfo->use_color && colorBuffBits) ? colorBuffBits : monoBuffBits);
-      SetBkMode(bufferDC, TRANSPARENT);
-      SetBkColor(bufferDC, get_color(gwinfo->backColor));
-      SetTextColor(bufferDC, get_color(gwinfo->drawColor));
-      SelectObject(bufferDC, hGraphFont);
-      SetROP2(bufferDC, gwinfo->drawMode ? R2_NOT : R2_COPYPEN);
-      SetPolyFillMode(bufferDC, WINDING);
-      SetViewportOrgEx(bufferDC, -gwinfo->view_h, -gwinfo->view_v, NULL);
-      ResetDCClipRect(bufferDC, gwinfo);
-      GetClientRect((HWND) gwinfo->window, &r);
-      r.left += gwinfo->view_h; r.top += gwinfo->view_v;
-      r.right += gwinfo->view_h; r.bottom += gwinfo->view_v;
-      hbr = GET_ERASE_BRUSH(gwinfo);
-      FillRect(bufferDC, &r, hbr);
-      RELEASE_BRUSH(hbr);
+      init_dc(gwinfo, bufferDC, TRUE);
       currentDC = bufferDC;
     }
     bufflevel++;
@@ -1215,7 +1266,138 @@ void StGWResetBuffer(void)
   }
 }
 
-void StGWDumpImage() { xlfail("not supported yet"); } // ### fix
+// ### fix
+void StGWDumpImage(StGWWinInfo *gwinfo, FILEP file, double scale)
+{
+  xlfail("not supported yet");
+}
+
+#ifdef WIN32
+static int GetPrinterDC(HDC *pdc)
+{
+  static PRINTDLG pd = { sizeof (PRINTDLG) } ;
+
+  /* Show the Print dialog box and get printer DC */
+  pd.Flags = PD_RETURNDC | PD_NOPAGENUMS | PD_NOSELECTION
+    | PD_USEDEVMODECOPIESANDCOLLATE;
+  if (!PrintDlg (&pd))
+    return FALSE;
+  else {
+    *pdc = pd.hDC;
+    return TRUE;
+  }
+}
+
+static int PrintEMF(HDC printerDC, HENHMETAFILE hemf)
+{
+  static DOCINFO di = { sizeof (DOCINFO), TEXT ("EmfView: Printing") } ;
+  ENHMETAHEADER emh;
+  int play_success, print_success, width, height, dwidth, dheight;
+  RECT r;
+  double hScale, vScale, scale;
+
+  /* get the metafile header */
+  GetEnhMetaFileHeader(hemf, sizeof(emh), &emh);
+  
+  /* Calculate the scale factor */
+  dwidth = GetDeviceCaps(printerDC, HORZRES);
+  width = emh.rclBounds.right - emh.rclBounds.left; 
+  if (width > 0) hScale = (double) dwidth / width;
+  else hScale = 1.0;
+  dheight = GetDeviceCaps(printerDC, VERTRES);
+  height = emh.rclBounds.bottom - emh.rclBounds.top;
+  if (height > 0) vScale = (double) dheight/ height;
+  else vScale = 1.0;
+  scale = hScale > vScale ? vScale : hScale;
+  if (scale > 1.0) scale = floor(scale);
+
+  /* Set the display rectangle */
+  /***** add margins? */
+  width = scale * width;
+  height = scale * height;
+  r.left = (dwidth - width) / 2;
+  r.top = (dheight - height) / 2;
+  r.right = r.left + width;
+  r.bottom = r.top + height;
+
+  /* Play the EMF to the printer */
+  /**** save current cursor and set up wait cursor */
+  print_success = FALSE;
+  if ((StartDoc(printerDC, &di) > 0) && (StartPage(printerDC) > 0)) {
+    play_success = PlayEnhMetaFile (printerDC, hemf, &r);
+    if (EndPage(printerDC) > 0) {
+      print_success = TRUE;
+      EndDoc(printerDC) ;
+    }
+  }
+  /**** restore original cursor */
+  return play_success;
+}
+#endif /* WIN32 */
+
+static void StGWPrint(StGWWinInfo *gwinfo)
+{
+#ifdef WIN32
+  /***** need to figure out how to disable for win32s */
+  if (gwinfo && gwinfo->window) {
+    HDC printerDC;
+    HENHMETAFILE hEMF;
+    int success;
+
+    /* Show the Print dialog box and get printer DC */
+    if (! GetPrinterDC(&printerDC))
+      return;
+    if (printerDC == NULL)
+      xlfail("Can't get printer DC");
+  
+    /* Draw the graph into an enhanced metafile */
+    {
+      CONTEXT cntxt, *target;
+      int jumping, mask;
+      LVAL val;
+
+      xlbegin(&cntxt,CF_UNWIND,NIL);
+      if (setjmp(cntxt.c_jmpbuf)) {
+	target = xltarget;
+	mask = xlmask;
+	val = xlvalue;
+	jumping = TRUE;
+      }
+      else {
+	target = NULL;
+	mask = 0;
+	val = NIL;
+	jumping = FALSE;
+
+	/* protected code */
+	if (buffering) xlfail("already buffering");
+	StGWResetBuffer();
+	bufferDC = CreateEnhMetaFile(NULL, NULL, NULL, NULL);
+	if (bufferDC == NULL) xlfail("can't create meta file");
+	init_dc(gwinfo, bufferDC, FALSE);
+	currentDC = bufferDC;
+	buffering = TRUE;
+	bufflevel++;
+	StGWObRedraw(GETWINOBJECT(gwinfo->window));
+      }
+      xlend(&cntxt);
+
+      /* cleanup code */
+      if (bufferDC != NULL) hEMF = CloseEnhMetaFile(bufferDC);
+      bufferDC = 0;
+      StGWResetBuffer();
+      if (jumping) DeleteDC(printerDC);
+      if (hEMF == NULL) xlfail("failed to close metafile properly");
+      if (jumping) xljump(target, mask, val);
+    }
+
+    success = PrintEMF(printerDC, hEMF);
+    DeleteDC (printerDC);
+    DeleteEnhMetaFile(hEMF);
+    if (! success) xlfail("printing failed");
+  }
+#endif /* WIN32 */
+}
 
 /**************************************************************************/
 /**                                                                      **/
@@ -1232,7 +1414,7 @@ void StGWDrawPoint(StGWWinInfo *gwinfo, int x, int y)
 
   if (gwinfo && (w = gwinfo->window) != 0) {
     oldWidth = gwinfo->drawPen.lopnWidth.x;
-    gwinfo->drawPen.lopnWidth.x = 0;
+    gwinfo->drawPen.lopnWidth.x = 1;
     hDC = GET_DC(w);
     oldPen = SET_PEN(hDC, gwinfo);
     MoveToEx(hDC, x, y, NULL);
@@ -1524,18 +1706,32 @@ static void init_msw_text(void)
   HFONT oldFont;
   HDC hDC;
   LOGFONT lf;
+  POINT pt;
 
   memset(&lf, 0, sizeof(LOGFONT));
-  GetPrivateProfileString(graphicsSection, "Font", "FIXEDSYS",
+  GetPrivateProfileString(graphicsSection, "Font", "Courier New",
                           lf.lfFaceName, LF_FACESIZE, iniFile);
-  lf.lfHeight = -GetPrivateProfileInt(graphicsSection,
-                                      "FontSize", 0, iniFile);
-  hGraphFont = CreateFontIndirect(&lf);
+  pt.y = -GetPrivateProfileInt(graphicsSection, "FontSize", 11, iniFile);
 
   hDC = GetDC(hWndFrame);
+  
+  /* convert specified point size to pixels */
+  pt.y = (pt.y * GetDeviceCaps(hDC, LOGPIXELSY)) / 72;
+  DPtoLP(hDC, &pt, 1);
+  lf.lfHeight = pt.y;
+
+  hGraphFont = CreateFontIndirect(&lf);
   oldFont = SelectObject(hDC, hGraphFont);
   GetTextMetrics(hDC, &tm);
   SelectObject(hDC, oldFont);
+
+  if (tm.tmPitchAndFamily & TMPF_TRUETYPE) {
+    /* rotate 90 degrees for drawing up */
+    lf.lfEscapement = lf.lfOrientation = 900;
+    hGraphFontUp = CreateFontIndirect(&lf);
+  }
+  else hGraphFontUp = NULL;
+
   charDC = CreateCompatibleDC(hDC);
   upcharDC = CreateCompatibleDC(hDC);
   ReleaseDC(hWndFrame, hDC);
@@ -1569,6 +1765,8 @@ static void cleanup_msw_text(void)
   }
   if (hGraphFont)
     DeleteObject(hGraphFont);
+  if (hGraphFontUp)
+    DeleteObject(hGraphFontUp);
 }
 
 static void draw_char(char ch, int x, int y, HDC hDC, int mode)
@@ -1587,22 +1785,42 @@ static void draw_char(char ch, int x, int y, HDC hDC, int mode)
 
 static void draw_char_up(char ch, int x, int y, HDC hDC, int mode)
 {
-  int i, j;
-  HDC targetDC;
-  COLORREF dc;
+  if (hGraphFontUp) {
+    HFONT hfontOld;
+    /* offsets of 1 seem to be needed to mach bitmap results */
+    x++;
+    y++;
+    if (mode) {
+      hfontOld = SelectObject(charDC, hGraphFontUp);
+      PatBlt(charDC, 0, 0, charSize, charSize, WHITENESS);
+      TextOut(charDC, 0, tm.tmAveCharWidth, &ch, 1);
+      BitBlt(hDC, x, y, charSize, charSize, charDC, 0, 0,  0x990066L);
+      SelectObject(charDC, hfontOld);
+    }
+    else {
+      hfontOld = SelectObject(hDC, hGraphFontUp);
+      draw_char(ch, x, y, hDC, mode);
+      SelectObject(hDC, hfontOld);
+    }
+  }
+  else {
+    int i, j;
+    HDC targetDC;
+    COLORREF dc;
 
-  targetDC = mode ? upcharDC : hDC;
-  dc = mode ? 0 : GetTextColor(hDC);
-  PatBlt(charDC, 0, 0, charSize, charSize, WHITENESS);
-  TextOut(charDC, 0, 0, &ch, 1);
-  if (mode)
-    PatBlt(upcharDC, 0, 0, charSize, charSize, WHITENESS);
-  for (i = 0; i < tm.tmAveCharWidth; i++)
-    for (j = 0; j < tm.tmHeight; j++)
-      if (GetPixel(charDC, i, j) == 0)
-	SetPixel(targetDC, x + j, y + tm.tmAveCharWidth - i, dc);
-  if (mode)
-    BitBlt(hDC, x, y, charSize, charSize, upcharDC, 0, 0,  0x990066L);
+    targetDC = mode ? upcharDC : hDC;
+    dc = mode ? 0 : GetTextColor(hDC);
+    PatBlt(charDC, 0, 0, charSize, charSize, WHITENESS);
+    TextOut(charDC, 0, 0, &ch, 1);
+    if (mode)
+      PatBlt(upcharDC, 0, 0, charSize, charSize, WHITENESS);
+    for (i = 0; i < tm.tmAveCharWidth; i++)
+      for (j = 0; j < tm.tmHeight; j++)
+	if (GetPixel(charDC, i, j) == 0)
+	  SetPixel(targetDC, x + j, y + tm.tmAveCharWidth - i, dc);
+    if (mode)
+      BitBlt(hDC, x, y, charSize, charSize, upcharDC, 0, 0,  0x990066L);
+  }
 }
 
 void StGWDrawString(StGWWinInfo *gwinfo, char *s, int x, int y)
@@ -1639,7 +1857,8 @@ void StGWDrawStringUp(StGWWinInfo *gwinfo, char *s, int x, int y)
   if (s && gwinfo && (w = gwinfo->window) != 0) {
     hDC = GET_DC(w);
     x -= tm.tmAscent;
-    y -= tm.tmAveCharWidth;
+    if (gwinfo->drawMode || hGraphFontUp == NULL)
+      y -= tm.tmAveCharWidth;
     if (gwinfo->drawMode) {
       SetBkColor(hDC, get_color(0));
       SetTextColor(hDC, get_color(1));
@@ -1666,7 +1885,7 @@ int StGWTextDescent(StGWWinInfo *gwinfo)
   return(tm.tmDescent);
 }
 
-StGWTextWidth(StGWWinInfo *gwinfo, char *text)
+int StGWTextWidth(StGWWinInfo *gwinfo, char *text)
 #pragma argsused text
 {
   return(text ? strlen(text) * tm.tmAveCharWidth : 0);
