@@ -1,7 +1,9 @@
 #include "xlisp.h"
 #include "xlstat.h"
 #include <ddeml.h>
-#include <alloc.h>
+#ifndef MinGW32
+#  include <alloc.h>
+#endif
 #include "ledit.h"
 #include "winutils.h"
 #include "wxlisp.h"
@@ -43,7 +45,7 @@ BOOL win32s;
 extern LVAL s_in_callback;
 
 static DWORD ddeInst = 0;
-static HSZ service, topic;
+static HSZ service, systopic;
 static jmp_buf exit_tag;
 
 /* forward declarations */
@@ -64,6 +66,8 @@ static void set_dde_result(HCONV, LVAL);
 static LVAL get_dde_result(HCONV);
 static void delete_dde_result(HCONV);
 
+int MSWAnyEventsQueued(void);
+
 #define LSGR_LINES 0
 #define LSGR_POINTS 1
 #define LSGR_MARGIN 10
@@ -92,14 +96,22 @@ int WINAPI WinMain(HINSTANCE hInstance,
   lpDdeProc = (PFNCALLBACK) MakeProcInstance((FARPROC) DdeProc, hInst);
   DdeInitialize(&ddeInst, lpDdeProc,
                 APPCLASS_STANDARD |
-                CBF_FAIL_ADVISES | CBF_FAIL_POKES |
-                CBF_SKIP_CONNECT_CONFIRMS |
+                CBF_FAIL_ADVISES |
+		/***** CBF_FAIL_POKES | CBF_SKIP_CONNECT_CONFIRMS |*/
                 CBF_SKIP_REGISTRATIONS | CBF_SKIP_UNREGISTRATIONS,
                 0L);
 
   service = DdeCreateStringHandle(ddeInst, "XLISP-STAT", CP_WINANSI);
-  topic = DdeCreateStringHandle(ddeInst, "XLISP-STAT", CP_WINANSI);
-  DdeNameService(ddeInst, service, NULL, DNS_REGISTER);
+  systopic = DdeCreateStringHandle(ddeInst, "SYSTEM", CP_WINANSI);
+  if (systopic == NULL) {
+    DdeFreeStringHandle(ddeInst, service);
+    service = NULL;
+  }
+  if (service != NULL)
+    if (! DdeNameService(ddeInst, service, NULL, DNS_REGISTER)) {
+      DdeFreeStringHandle(ddeInst, service);
+      service = NULL;
+    }
 
   /* parse the argument list into argv */
   {
@@ -152,10 +164,11 @@ int WINAPI WinMain(HINSTANCE hInstance,
   }
 
   /* drop DDEML */
-  DdeFreeStringHandle(ddeInst, service);
+  if (service != NULL) DdeFreeStringHandle(ddeInst, service);
+  if (systopic != NULL) DdeFreeStringHandle(ddeInst, systopic);
   DdeNameService(ddeInst, NULL, NULL, DNS_UNREGISTER);
   DdeUninitialize(ddeInst);
-  (void) FreeProcInstance((FARPROC) lpDdeProc);
+  FreeProcInstance((FARPROC) lpDdeProc);
 
   MSWCloseHelp();
   MSWGraphCleanup();
@@ -242,20 +255,12 @@ static BOOL InitInstance(HANDLE hInstance, int nCmdShow)
 {
   MDICREATESTRUCT mdicreate;
   RECT r;
-  LOGFONT lf;
 
   hInst = hInstance;
 
   InitInstDialogs(hInstance);
   InitInstGraphics(hInstance);
 
-  memset(&lf, 0, sizeof(LOGFONT));
-  GetPrivateProfileString(listenerSection, "Font", "FIXEDSYS",
-                          lf.lfFaceName, LF_FACESIZE, iniFile);
-  lf.lfHeight = -GetPrivateProfileInt(listenerSection,
-                                      "FontSize", 0, iniFile);
-  hListenerFont = CreateFontIndirect(&lf);
-
   if (GetPrivateProfileString(xlispSection, "Libdir", "", buf, 128, iniFile)) {
     int n;
     n = strlen(buf);
@@ -284,8 +289,6 @@ static BOOL InitInstance(HANDLE hInstance, int nCmdShow)
     }
   }
 
-  InitLEditClass(mainloop, hListenerFont);
-
   hAccel = LoadAccelerators(hInst, "MdiAccel");
 
   hMainMenu = LoadMenu(hInst, "MdiMenu");
@@ -315,6 +318,26 @@ static BOOL InitInstance(HANDLE hInstance, int nCmdShow)
   ShowWindow(hWndFrame, nCmdShow);
   UpdateWindow(hWndFrame);
 
+  {
+    LOGFONT lf;
+    HDC hDC;
+    POINT pt;
+    memset(&lf, 0, sizeof(LOGFONT));
+    GetPrivateProfileString(listenerSection, "Font", "Courier New",
+			    lf.lfFaceName, LF_FACESIZE, iniFile);
+    pt.y = -GetPrivateProfileInt(listenerSection, "FontSize", 11, iniFile);
+
+    /* convert specified point size to pixels */
+    hDC = GetDC(hWndFrame);
+    pt.y = (pt.y * GetDeviceCaps(hDC, LOGPIXELSY)) / 72;
+    DPtoLP(hDC, &pt, 1);
+    lf.lfHeight = pt.y;
+    ReleaseDC(hWndFrame, hDC);
+
+    hListenerFont = CreateFontIndirect(&lf);
+    InitLEditClass(mainloop, hListenerFont);
+  }
+
   /* create the listener frame */
   GetClientRect(hWndClient, (LPRECT) &r);
   mdicreate.szClass = szListenerClass;
@@ -325,14 +348,13 @@ static BOOL InitInstance(HANDLE hInstance, int nCmdShow)
   mdicreate.cx = r.right - r.left;
   mdicreate.cy = CW_USEDEFAULT;
   mdicreate.style = WS_MINIMIZE;
-  mdicreate.lParam = NULL;
+  mdicreate.lParam = 0;
   hEditFrame = MDICreateWindow(hWndClient, &mdicreate);
   if (! hEditFrame) {
     DestroyWindow(hWndClient);
     DestroyWindow(hWndFrame);
     return(FALSE);
   }
-  SetWindowStyle(hEditFrame, GetWindowStyle(hEditFrame) & ~WS_MAXIMIZEBOX);
 
   /* create the listener editor window pane */
   hEditWnd = CreateLEditWindow(hEditFrame, (HMENU) IDC_EDIT, hInst);
@@ -355,7 +377,6 @@ LONG CALLBACK FrameWndProc(HWND hWnd, UINT message, WPARAM wParam, LONG lParam)
 {
   static HWND hWndClient;
   WNDENUMPROC lpfnEnum;
-  DLGPROC lpProcAbout;
   HWND hWndChild;
   CLIENTCREATESTRUCT clientcreate;
 
@@ -390,6 +411,10 @@ LONG CALLBACK FrameWndProc(HWND hWnd, UINT message, WPARAM wParam, LONG lParam)
               );
       MessageBox(NULL, buf, "XLISP-STAT", MB_TASKMODAL);
       break;
+    case IDM_PRINT:
+      hWndChild = MDIGetActiveWindow(hWndClient);
+      SendMessage(hWndChild, WM_COMMAND, wParam, lParam);
+      return 0;
     case IDM_CLOSE:
       hWndChild = MDIGetActiveWindow(hWndClient);
       SendMessage(hWndChild, WM_CLOSE, 0, 0);
@@ -409,7 +434,7 @@ LONG CALLBACK FrameWndProc(HWND hWnd, UINT message, WPARAM wParam, LONG lParam)
     case IDM_CLOSEALL:
       lpfnEnum = (WNDENUMPROC) MakeProcInstance((FARPROC) CloseEnumProc, hInst);
       EnumChildWindows(hWndClient, lpfnEnum, 0);
-      (void) FreeProcInstance((FARPROC) lpfnEnum);
+      FreeProcInstance((FARPROC) lpfnEnum);
       return(0);
     default:
       if (IsLispMenuItem(GET_WM_COMMAND_ID(wParam, lParam))) {
@@ -443,81 +468,299 @@ LONG CALLBACK FrameWndProc(HWND hWnd, UINT message, WPARAM wParam, LONG lParam)
   return(DefFrameProc(hWnd, hWndClient, message, wParam, lParam));
 }
 
+static LVAL StringHandle2LispString(HSZ hsz)
+{
+  if (hsz == NULL)
+    return NIL;
+  else {
+    LVAL val;
+    DdeQueryString(ddeInst, hsz, buf, STRMAX, CP_WINANSI);
+    xlsave1(val);
+    val = cvstring(buf);
+    val = xlcallsubr1(xnupcase, val);
+    xlpop();
+    return val;
+  }
+}
+
+static LVAL DdeData2LispString(HDDEDATA hdata, int binary)
+{
+  LVAL val;
+  if (hdata != NULL) {
+    DWORD size;
+    char *p = (char *) DdeAccessData(hdata, &size);
+    if (p != NULL) {
+      if (! binary) {
+	char *q = memchr(p, 0, size);
+	if (q != NULL) size = q - p;
+      }
+      val = newstring(size);
+      memcpy(getstring(val), p, size);
+    }
+    else val = NIL;
+    DdeUnaccessData(hdata);
+    DdeFreeDataHandle(hdata);
+  }
+  else val = NIL;
+  return val;
+}
+
 HDDEDATA CALLBACK DdeProc(UINT type, UINT fmt, HCONV hconv,
                           HSZ hsz1, HSZ hsz2, HDDEDATA hdata,
                           DWORD dw1, DWORD dw2)
-#pragma argsused type fmt hconv hsz1 hsz2
 {
-  switch (type) {
-  case XTYP_CONNECT:
-    if (! DdeCmpStringHandles(topic, hsz1) &&
-        ! DdeCmpStringHandles(service, hsz2))
-      return (HDDEDATA) TRUE;
-    else
-      return (HDDEDATA) FALSE;
-  case XTYP_EXECUTE:
+  static int inited = FALSE;
+  static LVAL callbacksym;
+  static LVAL k_connect, k_connect_confirm, k_wildconnect, k_disconnect;
+  static LVAL k_execute, k_request, k_poke, k_text;
+
+  if (! inited) {
+    callbacksym = xlenter("SYSTEM::DDE-SERVER-CALLBACK");
+    k_connect = xlenter(":CONNECT");
+    k_connect_confirm = xlenter(":CONNECT-CONFIRM");
+    k_wildconnect = xlenter(":WILDCONNECT");
+    k_disconnect = xlenter(":DISCONNECT");
+    k_execute = xlenter(":EXECUTE");
+    k_request = xlenter(":REQUEST");
+    k_poke = xlenter(":POKE");
+    k_text = xlenter(":TEXT");
+    inited = TRUE;
+  }
+
+  if (fboundp(callbacksym)) {
+    LVAL typesym, fmtsym;
+    LVAL val = NIL;
+    int success;
+    int binary = TRUE;
+
+    /* Standardize the arguments so they can be converted to Lisp.
+       In particular, for transactions that do not use hdata, hdata is
+       set to NULL in case Windows passes something else in the hope
+       that it won't be used. */
+    switch (type) {
+    case XTYP_CONNECT:
+      typesym = k_connect;
+      fmt = 0;
+      hdata = NULL;
+      break;
+    case XTYP_CONNECT_CONFIRM:
+      typesym = k_connect_confirm;
+      fmt = 0;
+      hdata = NULL;
+      break;
+    case XTYP_WILDCONNECT:
+      typesym = k_wildconnect;
+      fmt = 0;
+      hdata = NULL;
+      break;
+    case XTYP_DISCONNECT:
+      typesym = k_disconnect;
+      fmt = 0;
+      hsz1 = NULL;
+      hsz2 = NULL;
+      hdata = NULL;
+      break;
+    case XTYP_EXECUTE:
+      typesym = k_execute;
+      fmt = 0;
+      hsz2 = NULL;
+      binary = FALSE;
+      break;
+    case XTYP_REQUEST:
+      typesym = k_request;
+      hdata = NULL;
+      break;
+    case XTYP_POKE:
+      typesym = k_poke;
+      break;
+    default: typesym = NIL;
+    }
+    fmtsym = fmt ==  CF_TEXT ? k_text : NIL;
+
+    /* Call the Lisp callback function.  This code traps any
+       non-local exit attempts -- the unwind stops here. I think
+       this is safe as far as XLISP is concerned; it is necessary
+       for self sends to work properly.  */
     {
       CONTEXT cntxt;
-      char *p = (char *) DdeAccessData(hdata, NULL);
-
-      /**** check for proper topic */
-      /**** check for format doesn't work -- seems to pass 0 */
-
-      /* This code traps any non-local exit attempts -- the unwind stops */
-      /* here. I think this is safe as far as XLISP is concerned; it is  */
-      /* necessary for self sends to work properly.                      */
       xlbegin(&cntxt, CF_UNWIND | CF_ERROR, NIL);
-      if (! setjmp(cntxt.c_jmpbuf)) {
-        LVAL olddenv = xldenv;
-        xldbind(s_breakenable, NIL);
-        set_dde_result(hconv, readevalstream(string2stream(p)));
-        xlunbind(olddenv);
+      if (setjmp(cntxt.c_jmpbuf))
+	success = FALSE;
+      else {
+	FRAMEP newfp;
+	int argc;
+	LVAL olddenv = xldenv;
+	
+	xldbind(s_breakenable, NIL);
+	newfp = xlsp;
+	pusharg(cvfixnum((FIXTYPE)(newfp - xlfp)));
+	pusharg(callbacksym);
+	pusharg(NIL);  /* will become argc */
+	argc = 0;
+	pusharg(typesym); argc++;
+	pusharg(fmtsym); argc++;
+	pusharg(cvfixnum((FIXTYPE) hconv)); argc++;
+	pusharg(StringHandle2LispString(hsz1)); argc++;
+	pusharg(StringHandle2LispString(hsz2)); argc++;
+	pusharg(DdeData2LispString(hdata, binary)); argc++;
+	pusharg(cvfixnum((FIXTYPE) dw1)); argc++;
+	pusharg(cvfixnum((FIXTYPE) dw2)); argc++;
+	newfp[2] = cvfixnum((FIXTYPE)argc);
+	xlfp = newfp;
+	val = xlapply(argc);
+	xlunbind(olddenv);
+	success = TRUE;
       }
       xlend(&cntxt);
-
-      DdeUnaccessData(hdata);
-      return (HDDEDATA) DDE_FACK;
     }
-  case XTYP_DISCONNECT:
-    delete_dde_result(hconv);
-    return NULL;
-  case XTYP_REQUEST:
-    {
-      CONTEXT cntxt;
-      LVAL val, stream;
-      HDDEDATA result = NULL;
 
-      /**** check for proper topic */
-      DdeQueryString(ddeInst, hsz2, buf, STRMAX, CP_WINANSI);
-      if (stricmp(buf, "VALUE"))
-        return NULL;
+    /* Build the result.  What is expected by DDE depends on the
+       transaction type. */
+    switch (type) {
+    case XTYP_CONNECT:
+      return (HDDEDATA) ((success && ! null(val)) ? TRUE : FALSE);
+    case XTYP_CONNECT_CONFIRM:
+      return NULL;
+    case XTYP_WILDCONNECT:
+      if (success) {
+	LVAL tmp;
+	HDDEDATA data = NULL;
+	int n;
+	for (n = 0, tmp = val;
+	     consp(tmp) && consp(car(tmp)) && stringp(car(car(tmp))) &&
+	       stringp(car(cdr(car(tmp))));
+	     n++, tmp = cdr(tmp));
+	if (n > 0) {
+	  int offset;
+	  HSZPAIR p;
+	  data = DdeCreateDataHandle(ddeInst, NULL, (n+1) * sizeof(HSZPAIR),
+				     0, NULL, 0, 0);
+	  for (tmp = val, offset = 0; n-- > 0; tmp = cdr(tmp)) {
+	    char *service = getstring(car(car(tmp)));
+	    char *topic = getstring(car(cdr(car(tmp))));
+	    p.hszSvc =  DdeCreateStringHandle(ddeInst, service, CP_WINANSI);
+	    p.hszTopic =  DdeCreateStringHandle(ddeInst, topic, CP_WINANSI);
+	    DdeAddData(data, (LPBYTE) &p, sizeof(p), offset);
+	    offset += sizeof(p);
+	  }
+	  p.hszSvc = p.hszTopic = NULL;
+	  DdeAddData(data, (LPBYTE) &p, sizeof(p), offset);
+	}
+	return data;
+      }
+      else return NULL;
+    case XTYP_EXECUTE:
+      if (null(val)) success = FALSE;
+      return (HDDEDATA) (success ? DDE_FACK : DDE_FNOTPROCESSED);
+    case XTYP_DISCONNECT:
+      return NULL;
+    case XTYP_REQUEST:
+      if (fmt == CF_TEXT && success && stringp(val)) {
+	unsigned char *p = (unsigned char *) getstring(val);
+	int size = getslength(val) + 1;
+	return DdeCreateDataHandle(ddeInst, p, size, 0, hsz2, CF_TEXT, 0);
+      }
+      else return NULL;
+    case XTYP_POKE:
+      if (null(val)) success = FALSE;
+      return (HDDEDATA) (success ? DDE_FACK : DDE_FNOTPROCESSED);
+    default:
+      return NULL;
+    }
+  }
+  else {
+    switch (type) {
+    case XTYP_CONNECT:
+      if (! DdeCmpStringHandles(service, hsz2) &&
+	  (! DdeCmpStringHandles(service, hsz1) ||
+	   ! DdeCmpStringHandles(systopic, hsz1)))
+	return (HDDEDATA) TRUE;
+      else return FALSE;
+    case XTYP_CONNECT_CONFIRM:
+      return NULL;
+    case XTYP_WILDCONNECT:
+      {
+	HSZPAIR p[3];
+	p[0].hszSvc = service;
+	p[0].hszTopic = service;
+	p[1].hszSvc = service;
+	p[1].hszTopic = systopic;
+	p[2].hszSvc = NULL;
+	p[2].hszTopic = NULL;
+	return DdeCreateDataHandle(ddeInst, (LPBYTE) p, sizeof(p),
+				   0, NULL, 0, 0);
+      }
+    case XTYP_EXECUTE:
+      if (! DdeCmpStringHandles(service, hsz1) ||
+	  ! DdeCmpStringHandles(systopic, hsz1)) {
+	int success;
+	char *p = (char *) DdeAccessData(hdata, NULL);
+	CONTEXT cntxt;
+	/* This code traps any non-local exit attempts -- the unwind
+	 stops here. I think this is safe as far as XLISP is
+	 concerned; it is necessary for self sends to work properly.  */
+	xlbegin(&cntxt, CF_UNWIND | CF_ERROR, NIL);
+	if (setjmp(cntxt.c_jmpbuf))
+	  success = FALSE;
+	else {
+	  LVAL olddenv = xldenv;
+	  xldbind(s_breakenable, NIL);
+	  set_dde_result(hconv, readevalstream(string2stream(p)));
+	  xlunbind(olddenv);
+	  success = TRUE;
+	}
+	xlend(&cntxt);
+	DdeUnaccessData(hdata);
+	DdeFreeDataHandle(hdata);
+	return (HDDEDATA) (success ? DDE_FACK : DDE_FNOTPROCESSED);
+      }
+      else return DDE_FNOTPROCESSED;
+    case XTYP_DISCONNECT:
+      delete_dde_result(hconv);
+      return NULL;
+    case XTYP_REQUEST:
       if (fmt != CF_TEXT)
-        return NULL;
+	return NULL;
+      else if (! DdeCmpStringHandles(service, hsz1) ||
+	       ! DdeCmpStringHandles(systopic, hsz1)) {
+	CONTEXT cntxt;
+	LVAL val, stream;
+	HDDEDATA result;
 
-      /* This code traps any non-local exit attempts -- the unwind stops */
-      /* here. I think this is safe as far as XLISP is concerned; it is  */
-      /* necessary for self sends to work properly.                      */
-      xlbegin(&cntxt, CF_UNWIND | CF_ERROR, NIL);
-      if (! setjmp(cntxt.c_jmpbuf)) {
-        LVAL olddenv = xldenv;
-        xldbind(s_breakenable, NIL);
-        val = get_dde_result(hconv);
-        xlsave1(stream);
-        stream = newustream();
-        xlcallsubr2(xprin1, val, stream);
-        stream = xlcallsubr1(xgetstroutput, stream);
-        result = DdeCreateDataHandle(ddeInst,
-                                     (unsigned char *) getstring(stream),
-                                     getslength(stream) + 1, 0,
-                                     hsz2, CF_TEXT, 0);
-        xlpop();
-        xlunbind(olddenv);
+	DdeQueryString(ddeInst, hsz2, buf, STRMAX, CP_WINANSI);
+	if (stricmp(buf, "VALUE"))
+	  return NULL;
+
+	/* This code traps any non-local exit attempts -- the unwind
+	   stops here. I think this is safe as far as XLISP is
+	   concerned; it is necessary for self sends to work properly.  */
+	xlbegin(&cntxt, CF_UNWIND | CF_ERROR, NIL);
+	if (setjmp(cntxt.c_jmpbuf))
+	  result = NULL;
+	else {
+	  LVAL olddenv = xldenv;
+	  xldbind(s_breakenable, NIL);
+	  val = get_dde_result(hconv);
+	  xlsave1(stream);
+	  stream = newustream();
+	  xlcallsubr2(xprin1, val, stream);
+	  stream = xlcallsubr1(xgetstroutput, stream);
+	  result = DdeCreateDataHandle(ddeInst,
+				       (unsigned char *) getstring(stream),
+				       getslength(stream) + 1, 0,
+				       hsz2, CF_TEXT, 0);
+	  xlpop();
+	  xlunbind(olddenv);
+	}
+	xlend(&cntxt);
+	return result;
       }
-      xlend(&cntxt);
-      return result;
+      else return NULL;
+    case XTYP_POKE:
+      return DDE_FNOTPROCESSED;
+    default:
+      return NULL;
     }
-  default:
-    return NULL;
   }
 }
 
@@ -543,11 +786,6 @@ LONG CALLBACK ListenerWndProc(HWND hWnd, UINT message, WPARAM wParam, LONG lPara
     return(0);
   case WM_SIZE:
     MoveWindow(hEditWnd, 0, 0, LOWORD(lParam), HIWORD(lParam), TRUE);
-    if (! IsIconic(hWnd)) { // ### to avoid problems with maximizing
-      DefMDIChildProc(hWnd, message, wParam, lParam);
-      MDIRestoreWindow(hWndClient, hWnd);
-      return(0);
-    }
     break;
   case WM_SETFOCUS:
     SetFocus(hEditWnd);
@@ -671,6 +909,48 @@ char *GetStringResource(int id)
   return gsr_buffer;
 }
 
+static HSZ Lisp2StringHandle(LVAL str)
+{
+  if (null(str))
+    return NULL;
+  else {
+    HSZ hsz;
+    char *p;
+    if (! stringp(str)) xlbadtype(str);
+    p = getstring(str);
+    hsz = DdeCreateStringHandle(ddeInst, p, CP_WINANSI);
+    if (hsz == NULL && p[0] != 0)
+      xlfail("string handle creation failed");
+    return hsz;
+  }
+}
+  
+#ifndef xlgastrornil
+#define xlgastrornil() (testarg(null(*xlargv) ? nextarg() : typearg(stringp)))
+#endif
+
+LVAL dde_name_service(void)
+{
+  LVAL servarg;
+  HSZ service;
+  int flag, success;
+
+  servarg = xlgastrornil();
+  flag = (! moreargs()) || (! null(xlgetarg())) ? 
+    DNS_REGISTER : DNS_UNREGISTER;
+  xllastarg();
+
+  service = Lisp2StringHandle(servarg);
+  success = DdeNameService(ddeInst, service, NULL, flag) ? TRUE : FALSE;
+  if (service != NULL) DdeFreeStringHandle(ddeInst, service);
+  return success ? s_true : NIL;
+}
+
+
+/***
+ *** DDE Client Interface
+ ***/
+
 #define MAXCONV 30
 
 static HCONV hconvs[MAXCONV] = { NULL };
@@ -704,22 +984,76 @@ static void close_all_conversations(void)
   }
 }
 
+LVAL dde_services(void)
+{
+  LVAL arg1, arg2, val, entry, string;
+  HSZ service = NULL, topic = NULL;
+  HCONVLIST hcl;
+  HCONV hconv;
+  CONVINFO ci;
+  CONTEXT cntxt,*target;
+  int mask, jumping;
+
+  arg1 = moreargs() ? xlgastrornil() : NIL;
+  arg2 = moreargs() ? xlgastrornil() : NIL;
+  xllastarg();
+
+  xlbegin(&cntxt,CF_UNWIND,NIL);
+  if (setjmp(cntxt.c_jmpbuf)) {
+    jumping = TRUE;
+    target = xltarget;
+    mask = xlmask;
+    val = xlvalue;
+  }
+  else {
+    jumping = FALSE;
+    service = Lisp2StringHandle(arg1);
+    topic = Lisp2StringHandle(arg2);
+
+    xlstkcheck(3);
+    xlsave(val);
+    xlsave(entry);
+    xlsave(string);
+    hconv = NULL;
+    val = NIL;
+    hcl = DdeConnectList(ddeInst, service, topic, 0, NULL);
+    while ((hconv = DdeQueryNextServer(hcl, hconv)) != NULL) {
+      DdeQueryConvInfo(hconv, QID_SYNC, &ci);
+      string = StringHandle2LispString(ci.hszTopic);
+      entry = consa(string);
+      string = StringHandle2LispString(ci.hszSvcPartner);
+      entry = cons(string, entry);
+      val = cons(entry, val);    
+    }
+    xlpopn(3);
+  }
+  xlend(&cntxt);
+
+  if (hcl != NULL) DdeDisconnectList(hcl);
+  if (service != NULL) DdeFreeStringHandle(ddeInst, service);
+  if (topic != NULL) DdeFreeStringHandle(ddeInst, topic);
+
+  if (jumping) xljump(target,mask,val);
+  
+  return xlnreverse(val);
+}
+
 LVAL dde_connect(void)
 {
-  char *servicename, *topicname;
+  LVAL servarg, toparg;
   HSZ service, topic;
   HCONV hconv;
   int hcd;
 
-  servicename = getstring(xlgastring());
-  topicname = moreargs() ? getstring(xlgastring()) : servicename;
+  servarg = moreargs() ? xlgastrornil() : NIL;
+  toparg = moreargs() ? xlgastrornil() : servarg;
   xllastarg();
 
   if ((hcd = find_conv_index()) < 0)
     xlfail("too many conversations in progress");
 
-  service = DdeCreateStringHandle(ddeInst, servicename, CP_WINANSI);
-  topic = DdeCreateStringHandle(ddeInst, topicname, CP_WINANSI);
+  service = Lisp2StringHandle(servarg);
+  topic = Lisp2StringHandle(toparg);
   hconv = DdeConnect(ddeInst, service, topic, NULL);
   DdeFreeStringHandle(ddeInst, service);
   DdeFreeStringHandle(ddeInst, topic);
@@ -736,83 +1070,90 @@ LVAL dde_client_transaction(void)
 {
   LVAL tmp;
   char *data, *item;
-  int hcd;
+  int hcd, binary;
   UINT type;
   DWORD timeout;
+  static int inited = FALSE;
+  static LVAL k_execute, k_poke, k_binary;
+
+  if (! inited) {
+    k_execute = xlenter(":EXECUTE");
+    k_poke = xlenter(":POKE");
+    k_binary = xlenter(":BINARY");
+    inited = TRUE;
+  }
 
   hcd = get_conversation_index();
+
   if (xlgetkeyarg(k_data, &tmp) && stringp(tmp))
     data = getstring(tmp);
   else
     data = NULL;
-  if (xlgetkeyarg(k_type, &tmp) && tmp == k_request)
-    type = XTYP_REQUEST;
+
+  if (xlgetkeyarg(k_type, &tmp)) {
+    if (tmp == k_request)
+      type = XTYP_REQUEST;
+    else if (tmp == k_execute)
+      type = XTYP_EXECUTE;
+    else if (tmp == k_poke)
+      type = XTYP_POKE;
+    else
+      xlbadtype(tmp);
+  }
   else
     type = XTYP_EXECUTE;
+
   if (xlgetkeyarg(k_item, &tmp) && stringp(tmp))
     item = getstring(tmp);
   else
     item = NULL;
+
   if (xlgetkeyarg(k_timeout, &tmp) && fixp(tmp) && getfixnum(tmp) > 0)
     timeout = getfixnum(tmp);
   else
     timeout = 60000L;
+
+  if (xlgetkeyarg(k_binary, &tmp) && tmp != NIL)
+    binary = TRUE;
+  else
+    binary = FALSE;
+
   xllastkey();
 
   if (hcd >= 0) {
-    switch (type) {
-    case XTYP_EXECUTE:
-      if (data == NULL)
-        return NIL;
-      else if (DdeClientTransaction((LPVOID) data, strlen(data) + 1,
-                                    hconvs[hcd], NULL, CF_TEXT, type,
-                                    timeout, NULL))
-        return s_true;
-      else
-        return NIL;
-    case XTYP_REQUEST:
-      {
-        CONTEXT cntxt,*target;
-        int mask,sts;
-        LVAL val;
-        HSZ hsz;
-        HDDEDATA hdata;
+    CONTEXT cntxt,*target;
+    int mask, jumping;
+    LVAL val;
+    HSZ hsz;
+    HDDEDATA hdata;
+    DWORD dsize = data != NULL ? strlen(data) + 1 : 0;
+    
+    if (item != NULL) hsz = DdeCreateStringHandle(ddeInst, item, CP_WINANSI);
+    else hsz = NULL;
 
-        if (item != NULL)
-          hsz = DdeCreateStringHandle(ddeInst, item, CP_WINANSI);
-        else
-          hsz = NULL;
-        xlbegin(&cntxt,CF_UNWIND,NIL);
-        if ((sts = setjmp(cntxt.c_jmpbuf)) != 0) {
-          target = xltarget;
-          mask = xlmask;
-          val = xlvalue;
-        }
-        else {
-          if ((hdata = DdeClientTransaction(NULL, 0, hconvs[hcd], hsz, CF_TEXT,
-                                            type, timeout, NULL)) != NULL)
-            val = cvstring((char *) DdeAccessData(hdata, NULL));
-          else
-            val = NIL;
-        }
-        xlend(&cntxt);
-
-        if (hsz != NULL)
-          DdeFreeStringHandle(ddeInst, hsz);
-        if (hdata != NULL)
-          DdeUnaccessData(hdata);
-
-        /* if unwinding, continue unwinding */
-        if (sts)
-	  xljump(target,mask,val);
-
-        return val;
-      }
-    /**** add POKE? */
-    /**** allow for asynchronous transactions? */
-    default:
-      return NIL;
+    xlbegin(&cntxt,CF_UNWIND,NIL);
+    if (setjmp(cntxt.c_jmpbuf)) {
+      jumping = TRUE;
+      target = xltarget;
+      mask = xlmask;
+      val = xlvalue;
     }
+    else {
+      jumping = FALSE;
+      hdata = DdeClientTransaction((LPVOID) data, dsize, hconvs[hcd], hsz,
+				   CF_TEXT, type, timeout, NULL);
+      if (type == XTYP_REQUEST)
+	val = DdeData2LispString(hdata, binary);
+      else val = hdata != NULL ? s_true : NIL;
+    }
+    xlend(&cntxt);
+
+    if (hsz != NULL) DdeFreeStringHandle(ddeInst, hsz);
+
+    /* if unwinding, continue unwinding */
+    if (jumping) xljump(target,mask,val);
+
+    return val;
   }
   else return NIL;
 }
@@ -848,9 +1189,10 @@ static void set_dde_result(HCONV hconv, LVAL val)
   LVAL s_dde_values = xlintern("*DDE-VALUES*", xlisppack);
   LVAL tmp1, tmp2, list;
 
-  xlstkcheck(2);
+  xlstkcheck(3);
   xlsave(tmp1);
   xlsave(tmp2);
+  xlprotect(val);
   tmp1 = cvfixnum((FIXTYPE) hconv);
   list = boundp(s_dde_values) ? getvalue(s_dde_values) : NIL;
   tmp2 = xlcallsubr2(xassoc, tmp1, list);
@@ -859,7 +1201,7 @@ static void set_dde_result(HCONV hconv, LVAL val)
     setvalue(s_dde_values, cons(tmp2, list));
   }
   rplacd(tmp2, val);
-  xlpopn(2);
+  xlpopn(3);
 }
 
 static LVAL get_dde_result(HCONV hconv)
@@ -870,6 +1212,7 @@ static LVAL get_dde_result(HCONV hconv)
   pair = xlcallsubr2(xassoc, cvfixnum((FIXTYPE) hconv), list);
   return consp(pair) ? cdr(pair) : NIL;
 }
+
 static void delete_dde_result(HCONV hconv)
 {
   LVAL s_dde_values = xlintern("*DDE-VALUES*", xlisppack);
@@ -881,6 +1224,11 @@ static void delete_dde_result(HCONV hconv)
   setvalue(s_dde_values, xlcallsubr2(xdelete, pair, list));
 }
 
+
+/***
+ *** User Profile Interface
+ ***/
+
 #define stringornil(x) (stringp(x) || null(x))
 #define xlgastringornil() (testarg(typearg(stringornil)))
 
