@@ -700,6 +700,16 @@ static VOID add_method P3C(LVAL, x, LVAL, selector, LVAL, method)
 
   check_object(x);
   if (! symbolp(selector)) xlerror("not a symbol", selector);
+  switch (ntype(method)) {
+  case BCCLOSURE:
+    if (getbcname(getbcccode(method)) == NIL)
+      setbcname(getbcccode(method), selector);
+    break;
+  case CLOSURE:
+    if (getname(method) == NIL)
+      setname(method, selector);
+    break;
+  }
   method_entry = find_own_method(x, selector);
   if (method_entry != NIL)
     set_method_entry_method(method_entry, method);
@@ -1250,11 +1260,6 @@ LVAL xsobject_isnew(V)
 VOID xsaddmsg P2C(LVAL, object, char *, str)
 {
   LVAL fcn;
-#ifdef _Windows
-  extern FUNDEF far funtab[];
-#else
-  extern FUNDEF funtab[];
-#endif
   static offset = FIRST_METHOD_OFFSET;
 
   xlsave1(fcn);
