@@ -32,8 +32,9 @@ extern LVAL xsabout_xlisp_stat(V);
 extern LVAL msw_cut(V), msw_copy(V), msw_paste(V), msw_clear(V),
   msw_copy_paste(V), msw_tile(V), msw_cascade(V), msw_closeall(V),
   msw_arrange_icons(V), msw_exit(V), msw_win_exec(V), msw_win_help(V),
-  msw_free_mem(V), msw_cursor_size(V);
+  msw_free_mem(V), msw_cursor_size(V), msw_print(V), msw_pagesetup(V);
 extern LVAL dde_connect(V), dde_client_transaction(V), dde_disconnect(V);
+extern LVAL dde_name_service(V), dde_services(V);
 extern LVAL xsopenfiledialog(V), xssetfiledialog(V);
 extern LVAL msw_get_profile_string(V), msw_write_profile_string(V);
 #endif /* _Windows */
