@@ -461,6 +461,8 @@
 {	"MSW-WIN-HELP",			S, msw_win_help			},
 {	"OPEN-FILE-DIALOG",		S, xsopenfiledialog		},
 {	"SET-FILE-DIALOG",		S, xssetfiledialog		},
+{	"SYSTEM::DDE-NAME-SERVICE",	S, dde_name_service		},
+{	"DDE-SERVICES",			S, dde_services			},
 {	"DDE-CONNECT",			S, dde_connect			},
 {	"DDE-CLIENT-TRANSACTION",	S, dde_client_transaction	},
 {	"DDE-DISCONNECT",		S, dde_disconnect		},
@@ -851,6 +853,10 @@
 {	"UNCMIN-MODELHESS",	S,	xsminmodelhess		},
 {	"UNCMIN-GRADSIZE",	S,	xsmingradsize		},
 {	"UNCMIN-LINESEARCH",	S,	xsminlinesearch		},
+#ifdef _Windows
+{	"MSW-PRINT",		S,	msw_print		},
+{	"MSW-PAGESETUP",	S,	msw_pagesetup		},
+#endif
 #else
 {   "SYSTEM",   S,  xsystem },
 
