@@ -36,11 +36,6 @@ LOCAL LVAL set_cps_node_internals P1H(int);
 /*#define PROFILE*/
 #ifdef PROFILE
 LVAL s_profile_output;
-#ifdef _Windows   /* L. Tierney */
-extern FUNDEF far funtab[];
-#else
-extern FUNDEF funtab[];
-#endif
 #endif /* PROFILE */
 
 /* external variables and functions */
@@ -232,6 +227,12 @@ CONTINUATIONP xlcstop;
   xlcstop++; \
 }
 
+#define check_catch_continuation() \
+  if (xlcstop >= xlcsend) xlabort("continuation stack overflow")
+
+#define save_catch_continuation() xlcstop->pe.pc = current_pc
+#define get_catch_continuation() xlcstop->pe.pc
+
 #define set_saved_continuation_pc(npc) (xlcstop[-1].pe.pc = (npc))
 
 /*****
@@ -408,7 +409,29 @@ enum OPCODES {
   SET_CAR_OP,
   SET_CDR_OP,
   INIT_0_OP,
-  STOP_OP
+  STOP_OP,
+  SWAP_OP,
+  LDCONST_OP,
+  NCASE_OP,
+  MKCLOS_OP,
+  SETCLOSDATA_OP,
+  SETCLOSCODE_OP,
+  LDNOTSUPP_OP,
+  LDMVARGS_OP,
+  NOT_OP,
+  NEW_BLOCK_OP,
+  NEW_TAGBODY_OP,
+  NEW_GO_OP,
+  NEW_CATCH_OP,
+  NEW_ERRSET_OP,
+  NEW_UNWIND_PROTECT_OP,
+  GET_OPTARG_OP,
+  MAKE_KEYARGS_OP,
+  CHECK_LAST_KEYARG_OP,
+  GET_KEYARG_OP,
+  NEW_DYNAMIC_BIND_OP,
+  NEW_DYNAMIC_UNBIND_OP,
+  STRUCT_OP
 };
 
 
@@ -599,6 +622,7 @@ LOCAL bytecode *do_ARITH1_OP P1H(bytecode *);
 LOCAL bytecode *do_AREF2_OP P2H(bytecode *, int);
 LOCAL bytecode *do_CXR_OP P1H(bytecode *);
 LOCAL bytecode *do_MAKE_Y_CLOSURES_OP P1H(bytecode *);
+LOCAL bytecode *do_MKCLOS_OP P1H(bytecode *);
 LOCAL bytecode *do_INIT_OP P1H(bytecode *);
 
 LOCAL bytecode *do_ARITH2_OP P1C(bytecode *, current_pc)
@@ -978,6 +1002,29 @@ LOCAL bytecode *do_MAKE_Y_CLOSURES_OP P1C(bytecode *, current_pc)
   return(current_pc);
 }
 
+LOCAL bytecode *do_MKCLOS_OP P1C(bytecode *, current_pc)
+{
+  int ri, n, nv, i;
+  LVAL env, closure;
+  
+  xlstkcheck(2);
+  xlsave(env);
+  xlsave(closure);
+  
+  get_operand(nv);
+  get_operand(n);
+  env = newvector(nv);
+  for (i = 0; i < n; i++) {
+    get_operand(ri);
+    closure = copy_bcode(current_function);
+    set_bcode_environment(closure, env);
+    setregval(ri, closure);
+  }
+  xlpopn(2);
+
+  return(current_pc);
+}
+
 /**** see if this can be cleaned up */
 LOCAL bytecode *do_INIT_OP P1C(bytecode *, current_pc)
 {
@@ -1125,6 +1172,21 @@ LOCAL bytecode *do_INIT_OP P1C(bytecode *, current_pc)
   return(current_pc);
 }
 
+static LVAL bc_get_keyarg P4C(LVAL, args, LVAL, key, LVAL, dflt, int *, pfound)
+{
+  LVAL last = args;
+  for (args = cdr(args); consp(args); last = cdr(args), args = cdr(last)) {
+    if (! consp(cdr(args))) xlfail("keyword value missing");
+    if (car(args) == key) {
+      rplacd(last, cdr(cdr(args)));
+      if (pfound != NULL) *pfound = TRUE;
+      return car(cdr(args));
+    }
+  }
+  if (pfound != NULL) *pfound = FALSE;
+  return dflt;
+}
+
 /* bcloop - the interpreter main loop */
 LOCAL VOID bcloop P1C(int, entry)
 {
@@ -1384,8 +1446,10 @@ LOCAL VOID bcloop P1C(int, entry)
 
 	get_operand(fi);
 	fun = getlitval(fi);
-	while (! fboundp(fun)) xlfunbound(fun);
-	fun = getfunction(fun);
+	if (symbolp(fun)) {
+	  while (! fboundp(fun)) xlfunbound(fun);
+	  fun = getfunction(fun);
+	}
       }
     save_call_body:
       {
@@ -1418,8 +1482,10 @@ LOCAL VOID bcloop P1C(int, entry)
 	
 	get_operand(fi);
 	fun = getlitval(fi);
-	while (! fboundp(fun)) xlfunbound(fun);
-	fun = getfunction(fun);
+	if (symbolp(fun)) {
+	  while (! fboundp(fun)) xlfunbound(fun);
+	  fun = getfunction(fun);
+	}
       }
     call_body:
       {
@@ -2028,6 +2094,8 @@ LOCAL VOID bcloop P1C(int, entry)
 	       consp(nexts) && consp(nextv);
 	       nexts = cdr(nexts), nextv = cdr(nextv))
 	    xldbind(car(nexts), car(nextv));
+	  for (; consp(nexts); nexts = cdr(nexts))
+	    xldbind(car(nexts), s_unbound);
 	}
       }
       break;
@@ -2185,6 +2253,430 @@ LOCAL VOID bcloop P1C(int, entry)
     case STOP_OP:
       goto done;
       break;
+    case SWAP_OP:
+      {
+	int xi, yi;
+	LVAL tmp;
+	get_operand(xi);
+	get_operand(yi);
+	tmp = getregval(yi);
+	setregval(yi, getregval(xi));
+	setregval(xi, tmp);
+      }
+      break;
+    case LDCONST_OP:
+      {
+	int xi, yi;
+	LVAL literals = bcode_literals(current_function);
+	get_operand(xi);
+	get_operand(yi);
+	setregval(yi, getlitval(xi));
+      }
+      break;
+    case NCASE_OP:
+      {
+	int ci, chi, li, i, n;
+	LVAL cl, chl, cs;
+	LVAL literals = bcode_literals(current_function);
+	
+	get_operand(ci);
+	get_operand(chi);
+	cl = getregval(ci);
+	chl = getlitval(chi);
+	n = getsize(chl);
+	
+	li = 0; /* to keep compiler happy */
+	for (i = 0; i < n; i++) {
+	  get_operand(li);
+	  cs = getelement(chl, i);
+	  if (case_match_p(cl, cs)) break;
+	}
+	if (i == n) get_operand(li);
+	
+	do_goto(li);
+      }
+      break;
+    case MKCLOS_OP:
+      current_pc = do_MKCLOS_OP(current_pc);
+      break;
+    case SETCLOSDATA_OP:
+      {
+	int ci, nv, xi, i;
+	LVAL env;
+	get_operand(ci);
+	env = bcode_environment(getregval(ci));
+	get_operand(nv); /**** could check against size of env */
+	for (i = 0; i < nv; i++) {
+	  get_operand(xi);
+	  setelement(env, i, getregval(xi));
+	}
+	break;
+      }
+    case SETCLOSCODE_OP:
+      {
+	int ci, fi, li;
+	LVAL closure;
+	LVAL literals = bcode_literals(current_function);
+	get_operand(ci);
+	get_operand(fi);
+	get_operand(li);
+	closure = getregval(ci);
+	set_bcode_index(closure, fi);
+	setbcname(getbcccode(closure), getlitval(li));
+	break;
+      }
+    case LDNOTSUPP_OP:
+      {
+	int vi;
+	get_operand(vi);
+	setregval(vi, s_not_supplied);
+	break;
+      }
+    case LDMVARGS_OP:
+      {
+	int nreq, nopt, restp, fi, vi, nro, i;
+	/*LVAL literals = bcode_literals(current_function);*/
+	get_operand(nreq);
+	get_operand(nopt);
+	get_operand(restp);
+	get_operand(fi);/***** currently ignored -- use for error messages */
+	nro = nreq + nopt;
+	if (nreq > xlnumresults) xltoofew();
+	if (! restp && xlnumresults > nro) xltoomany();
+	for (i = 0; i < nro && i < xlnumresults; i++) {
+	  get_operand(vi);
+	  setregval(vi, xlresults[i]);
+	}
+	for (i = xlnumresults; i < nro; i++) {
+	  get_operand(vi);
+	  setregval(vi, s_not_supplied);
+	}
+	if (restp) {
+	  get_operand(vi);
+	  setregval(vi, NIL);
+	  for (i = nro; i < xlnumresults; i++)
+	    setregval(vi, cons(xlresults[i], getregval(vi)));
+	  setregval(vi, xlnreverse(getregval(vi)));
+	}
+	break;
+      }
+    case NOT_OP:
+      {
+	int ai, vi;
+	get_operand(ai);
+	get_operand(vi);
+	setregval(vi, null(getregval(ai)) ? s_true: NIL);
+	break;
+      }
+    case NEW_BLOCK_OP:
+      {
+	volatile int ci;
+	LVAL * volatile oldbase = vsbase;
+	LVAL tag;
+	int bi;
+	CONTEXT cntxt;
+
+	get_operand(bi);
+	get_operand(ci);
+	xlsave1(tag);
+	tag = consa(NIL);
+	xlbegin(&cntxt, CF_THROW, tag);
+	if (! setjmp(cntxt.c_jmpbuf))
+	  xlapp1(getregval(bi), tag); /**** inefficient but easy */
+	/* vsbase needs to be restored in jumps. xlcstop is handled in
+           the standard jump code, as is vstop = xlsp */
+	vsbase = oldbase;
+	do_goto(ci);
+	xlend(&cntxt);
+	xlpop();
+	break;
+      }
+    case NEW_TAGBODY_OP:
+      {
+	volatile int ci;
+	LVAL * volatile oldbase = vsbase;
+	LVAL tag;
+	int bi;
+	CONTEXT cntxt;
+
+	get_operand(bi);
+	get_operand(ci);
+	xlsave1(tag);
+	tag = consa(NIL);
+	xlbegin(&cntxt, CF_THROW, tag);
+	if (setjmp(cntxt.c_jmpbuf)) {
+	  BC_evfun(xlvalue, 0, xlsp);
+	  /* vsbase needs to be restored in jumps. xlcstop is handled in
+	     the standard jump code, as is vstop = xlsp */
+	  vsbase = oldbase;
+	}
+	else {
+	  LVAL *oldsp = xlsp;
+	  pusharg(tag);
+	  BC_evfun(getregval(bi), 1, oldsp);
+	  xlsp = oldsp;
+	}
+	do_goto(ci);
+	xlend(&cntxt);
+	xlpop();
+	break;
+      }
+    case NEW_GO_OP:
+      {
+	int ri, li;
+	get_operand(ri);
+	get_operand(li);
+	xlthrow(getregval(ri), getregval(li));
+	break; /* not reached */
+      }
+    case NEW_CATCH_OP:
+      {
+	volatile int ci;
+	LVAL * volatile oldbase = vsbase;
+	int ti, bi;
+	CONTEXT cntxt;
+
+	get_operand(ti);
+	get_operand(bi);
+	get_operand(ci);
+	xlbegin(&cntxt, CF_THROW, getregval(ti));
+	if (! setjmp(cntxt.c_jmpbuf))
+	  BC_evfun(getregval(bi), 0, xlsp);
+	/* vsbase needs to be restored in jumps. xlcstop is handled in
+           the standard jump code, as is vstop = xlsp */
+	vsbase = oldbase;
+	do_goto(ci);
+	xlend(&cntxt);
+	break;
+      }
+    case NEW_ERRSET_OP:
+      {
+	volatile int ci;
+	LVAL * volatile oldbase = vsbase;
+	int bi, fi;
+	CONTEXT cntxt;
+
+	get_operand(bi);
+	get_operand(fi);
+	get_operand(ci);
+	xlbegin(&cntxt, CF_ERROR, getregval(fi));
+	if (setjmp(cntxt.c_jmpbuf)) {
+	  set_one_result(NIL);
+	  /* vsbase needs to be restored in jumps. xlcstop is handled in
+	     the standard jump code, as is vstop = xlsp */
+	  vsbase = oldbase;
+	}
+	else {
+	  LVAL val;
+	  BC_evfun(getregval(bi), 0, xlsp);
+	  val = get_one_result(); /* has to be done before set */
+	  set_one_result(consa(val));
+	}
+	do_goto(ci);
+	xlend(&cntxt);
+	break;
+      }
+    case NEW_UNWIND_PROTECT_OP:
+      {
+	volatile int ci, pi;
+	LVAL * volatile oldbase = vsbase;
+	int bi, sts = FALSE, nvals, i, mask;
+	CONTEXT cntxt, *target;
+	LVAL val;
+
+	get_operand(bi);
+	get_operand(pi);
+	get_operand(ci);
+	xlbegin(&cntxt,CF_UNWIND,NIL);
+	if (setjmp(cntxt.c_jmpbuf)) {
+	  sts = TRUE;
+	  target = xltarget;
+	  mask = xlmask;
+	  val = xlvalue;
+	  /* vsbase needs to be restored in jumps. xlcstop is handled in
+	     the standard jump code, as is vstop = xlsp */
+	  vsbase = oldbase;
+	}
+	else {
+	  BC_evfun(getregval(bi), 0, xlsp);
+	  nvals = xlnumresults;
+	  for (i = 0; i < nvals; i++) {
+	    val = get_nth_result(i);
+	    pusharg(val);
+	  }
+	}
+	xlend(&cntxt);
+	BC_evfun(getregval(pi), 0, xlsp);
+	if (sts)
+	  xljump(target, mask, val);
+	else {
+	  vstop -= nvals;
+	  for (i = 0; i < nvals; i++)
+	    set_nth_result(i, vstop[i]);
+	  xlnumresults = nvals;
+	  do_goto(ci);
+	}
+	break;
+      }
+    case GET_OPTARG_OP:
+      {
+	int ai, di, vi, si;
+	LVAL val;
+	get_operand(ai);
+	get_operand(di);
+	get_operand(vi);
+	get_operand(si);
+	val = getregval(ai);
+	if (val == s_not_supplied) {
+	  LVAL literals = bcode_literals(current_function);
+	  setregval(vi, getlitval(di));
+	  setregval(si, NIL);
+	}
+	else {
+	  setregval(vi, val);
+	  setregval(si, s_true);
+	}
+	break;
+      }
+    case MAKE_KEYARGS_OP:
+      {
+	int ai, vi;
+	get_operand(ai);
+	get_operand(vi);
+	setregval(vi, consd(copylist(getregval(ai))));
+	break;
+      }
+    case CHECK_LAST_KEYARG_OP:
+      {
+	int ai, others;
+	LVAL keys;
+	get_operand(ai);
+	get_operand(others);
+	keys = getregval(ai);
+	if (llength(cdr(keys)) % 2)
+	  xlfail("missing keyword value");
+	if (! null(cdr(keys)) &&
+	    ! others &&
+	    null(bc_get_keyarg(keys, k_allow_other_keys, NIL, NULL)) &&
+	    ! null(getvalue(s_strict_keywords)))
+	  xlfail("too many keyword arguments");
+	break;
+      }
+    case GET_KEYARG_OP:
+      {
+	int ai, ki, di, vi, si, found;
+	LVAL literals = bcode_literals(current_function);
+	LVAL val;
+	get_operand(ai);
+	get_operand(ki);
+	get_operand(di);
+	get_operand(vi);
+	get_operand(si);
+	val = bc_get_keyarg(getregval(ai), getlitval(ki), getlitval(di),
+			    &found);
+	setregval(vi, val);
+	setregval(si, found ? s_true : NIL);
+	break;
+      }
+    case NEW_DYNAMIC_BIND_OP:
+      {
+	int si, vi, ei;
+	LVAL nexts, nextv, olddenv = xldenv;
+	
+	get_operand(si);
+	get_operand(vi);
+	get_operand(ei);
+	
+	/**** For the moment, the old xldenv value is passed both in a
+              register and on the stack -- this is to help with
+              debugging. Eventually it will just be passed in a
+              register */
+	pusharg(xldenv);
+	
+	if (getregval(vi) == s_true) {
+	  for (nexts = getregval(si); consp(nexts); nexts = cdr(nexts))
+	    xldbind(car(nexts), getvalue(car(nexts)));
+	}
+	else {
+	  for (nexts = getregval(si), nextv = getregval(vi);
+	       consp(nexts) && consp(nextv);
+	       nexts = cdr(nexts), nextv = cdr(nextv))
+	    xldbind(car(nexts), car(nextv));
+	  for (; consp(nexts); nexts = cdr(nexts))
+	    xldbind(car(nexts), s_unbound);
+	}
+	setregval(ei, olddenv);
+	break;
+      }
+    case NEW_DYNAMIC_UNBIND_OP:
+      {
+	/**** For the moment, the old xldenv value is passed both in a
+              register and on the stack -- this is to help with
+              debugging. Eventually it will just be passed on the
+              stack */
+	LVAL olddenv = *--vstop;
+	int ei;
+	get_operand(ei);
+	if (getregval(ei) != olddenv)
+	  stdputstr("bad stack in unbind\n");
+	xlunbind(olddenv);
+	break;
+      }
+    case STRUCT_OP:
+      {
+	int si;
+	LVAL str;
+	get_operand(si);
+	str = getregval(si);
+	switch (next_opcode()) {
+	case 0:
+	  {
+	    int i, vi;
+	    get_operand(i);
+	    get_operand(vi);
+	    if (! structp(str)) xlbadtype(str);
+	    if (i >= getsize(str)) xlerror("bad structure reference",str);
+	    setregval(vi, getelement(str,i));
+	    break;
+	  }
+	case 1:
+	  {
+	    int i, ni, vi;
+	    LVAL val;
+	    get_operand(i);
+	    get_operand(ni);
+	    get_operand(vi);
+	    if (! structp(str)) xlbadtype(str);
+	    if (i >= getsize(str)) xlerror("bad structure reference",str);
+	    val = getregval(ni);
+	    setelement(str, i, val);
+	    setregval(vi, val);
+	    break;
+	  }
+	case 2:
+	  {
+	    int ti, vi, val = FALSE;
+	    LVAL literals = bcode_literals(current_function);
+	    LVAL type;
+	    get_operand(ti);
+	    get_operand(vi);
+	    type = getlitval(ti);
+	    if (structp(str)) {
+	      for (str = getelement(str,0);
+		   ! null(str);
+		   str = xlgetprop(str,s_strinclude))
+		if (str == type) {
+		  val = TRUE;
+		  break;
+		}
+	    }
+	    setregval(vi, val ? s_true : NIL);
+	    break;
+	  }
+	default: xlfail("unknown opcode");
+	}
+	break;
+      }
     default: xlfail("unknown opcode");
     }
   }
