@@ -202,6 +202,10 @@ LVAL xgetlambda(V)
     closure = xlgetarg();   /* fixed to allow operation when not a closure */
     if (closurep(closure)) 
       fun = cons(gettype(closure), cons(getlambda(closure), getbody(closure)));
+    else if (bcclosurep(closure)) {
+      LVAL def = getbcdef(getbcccode(closure));
+      fun = consp(def) ? car(def) : NIL;
+    }
     else
       fun = NIL;
 #ifdef MULVALS
@@ -211,6 +215,11 @@ LVAL xgetlambda(V)
       xlresults[1] = (! null(getenvi(closure))) ? s_true : NIL;
       xlresults[2] = getname(closure);
     }
+    else if (bcclosurep(closure)) {
+      LVAL def = getbcdef(getbcccode(closure));
+      xlresults[1] = consp(def) ? cdr(def) : s_true;
+      xlresults[2] = getbcname(getbcccode(closure));
+    }
     else {
       xlresults[1] = s_true;
       xlresults[2] = NIL;
@@ -2087,3 +2096,55 @@ LVAL xnthvalue(V)
   return((n >= 0 && n < (FIXTYPE)xlnumresults) ? xlresults[(int)n] : NIL);
 }  
 #endif /* MULVALS */
+
+/* xthe - special form 'the' */
+LVAL xthe(V)
+{
+  LVAL arg;
+  xlgetarg(); /* skip type argument */
+  arg = xlgetarg();
+  xllastarg();
+  return xleval(arg);
+}
+
+/* xsymaclet - special form symbol-macrolet */
+LVAL xsymaclet(V)
+{
+  LVAL binds;
+  LVAL newenv, val;
+
+  /* protect some pointers */
+  xlsave1(newenv);
+
+  /* create a new environment frame */
+  newenv = xlframe(xlenv);
+
+  /* get the list of bindings and bind the symbols */
+  for (binds = xlgalist(); consp(binds); binds = cdr(binds)) {
+    val = car(binds);
+    if (! consp(val) || ! symbolp(car(val)) ||
+	specialp(car(val)) || ! consp(cdr(val)))
+      xlfail("bad symbol macro binding");
+    xlpbind(k_symbol_macro, car(cdr(val)), newenv);
+    xlpbind(car(val), k_symbol_macro, newenv);
+  }
+  xlenv = newenv;
+
+  /* execute the code */
+#ifdef MULVALS
+  xlnumresults = 1;
+  xlresults[0] = NIL;
+#endif /* MULVALS */
+  for (val = NIL; moreargs(); )
+    val = xleval(nextarg());
+
+  /* unbind the arguments */
+  xlenv = cdr(xlenv);
+
+  /* restore the stack */
+  xlpop();
+
+  /* return the result */
+  return (val);
+}
+
