@@ -10,6 +10,7 @@ typedef LVAL (*subrfun)(V);
 static LVAL finalize_registered, finalize_pending;
 static VOID check_finalize(V);
 static VOID do_finalize(V);
+static VOID check_weak_boxes(V);
 
 #ifdef NEWGC
 
@@ -221,6 +222,8 @@ VOID xlminit(V)
     type_info[RNDSTATE].has_children = TRUE;
     type_info[NATPTR].allocated = FALSE;
     type_info[NATPTR].has_children = TRUE;
+    type_info[WEAKBOX].allocated = FALSE;
+    type_info[WEAKBOX].has_children = FALSE;
     type_info[DARRAY].allocated = FALSE;
     type_info[DARRAY].has_children = TRUE;
     type_info[STRUCT].has_children = TRUE;
@@ -1128,6 +1131,7 @@ LOCAL VOID ggc P1C(int, full_gc)
       unset_to_new_node(tmp);
     }
   }
+  check_weak_boxes();
 
   /* clean out free allocated nodes */
   for (sg = lastpermseg->sg_next; sg != NULL; sg = sg->sg_next) {
@@ -2925,7 +2929,7 @@ static VOID do_finalize(V)
   LVAL next;
 
   xlsave1(next);
-  xlbegin(&cntxt,CF_TOPLEVEL|CF_CLEANUP|CF_BRKLEVEL,(LVAL)1);
+  xlbegin(&cntxt,CF_UNWIND|CF_ERROR,NIL);
   setjmp(cntxt.c_jmpbuf);
   while (consp(finalize_pending)) {
     next = finalize_pending;
@@ -2936,6 +2940,29 @@ static VOID do_finalize(V)
   xlpop();
 }
 
+static void check_weak_boxes()
+{
+  LVAL last = NIL, next = xlweakboxes;
+
+  while (ntype(next) == WEAKBOX) {
+    if (is_new_node(next)) {
+      /* delete box from weak box list */
+      LVAL tail = cdr(next);
+      if (null(last))
+        xlweakboxes = tail;
+      else
+        Rplacd(last, tail);
+      next = tail;
+    }
+    else {
+      if (is_new_node(car(next)))
+	Rplaca(next, NIL);
+      last = next;
+      next = cdr(next);
+    }
+  }
+}
+
 LVAL xregfinal(V)
 {
   LVAL arg, fun;
@@ -2946,6 +2973,26 @@ LVAL xregfinal(V)
   return NIL;
 }
 
+LVAL xmkweakbox(V)
+{
+  LVAL val = xlgetarg();
+  LVAL box = consa(val);
+  xllastarg();
+  ntype(box) = WEAKBOX;
+  Rplacd(box, xlweakboxes);
+  xlweakboxes = box;
+  return box;
+}
+
+#define weakboxp(x) (ntype(x) == WEAKBOX)
+#define xlgaweakbox()    (testarg(typearg(weakboxp)))
+LVAL xweakboxval(V)
+{
+  LVAL box = xlgaweakbox();
+  xllastarg();
+  return car(box);
+}
+
 LVAL newnatptr P2C(ALLOCTYPE *, p, LVAL, v)
 {
   LVAL val;
