@@ -219,12 +219,16 @@ enum {
 #define getbclits(x)    getelement(x, 2)
 #define getbcidx(x)     getelement(x, 3)
 #define getbcenv(x)     getelement(x, 4)
+#define getbcname(x)    getelement(x, 5)
+#define getbcdef(x)     getelement(x, 6)
 #define setbccode(x, v) setelement(x, 0, v)
 #define setbcjtab(x, v) setelement(x, 1, v)
 #define setbclits(x, v) setelement(x, 2, v)
 #define setbcidx(x, v)  setelement(x, 3, v)
 #define setbcenv(x, v)  setelement(x, 4, v)
-#define BCODESIZE 5
+#define setbcname(x, v) setelement(x, 5, v)
+#define setbcdef(x, v)  setelement(x, 6, v)
+#define BCODESIZE 7
 #endif /* BYTECODE */
 
 #ifdef PACKAGES
@@ -259,18 +263,19 @@ enum {
 #define ADATA   10
 #define TVEC    11
 #define NATPTR  12  /* native pointer */
+#define WEAKBOX 13
 /* Non-array types from CONS up use CAR and CDR fields */
 /* This means that all types from CONS up have garbage collectable elements */
-#define CONS	13
-#define COMPLEX 14
+#define CONS	16
+#define COMPLEX 17
 #ifdef BIGNUMS
-#define RATIO	15
+#define RATIO	18
 #endif
-#define USTREAM 16
-#define DARRAY  17
-#define RNDSTATE 18
+#define USTREAM 19
+#define DARRAY  20
+#define RNDSTATE 21
 #ifdef BYTECODE
-#define BCCLOSURE 19
+#define BCCLOSURE 22
 #endif /* BYTECODE */
 #define ARRAY   32      /* arrayed types */
 #define SYMBOL  (ARRAY+1)
