@@ -1363,7 +1363,7 @@ LOCAL VOID write_double_ffmt P4C(char *, s, double, y, int, e, int, d)
 
   /* re-read exponent in case changed by rounding */
   if (read_exponent(strchr(s, 'e') + 1) > e) {
-    s[f + 2] = '0';         /* extend fractional part with a zero */
+    s[f > 0 ? f + 2 : 2] = '0'; /* extend fractional part with a zero */
     f++;                    /* increment length of fractional part */
     e++;                    /* increment exponent */
   }
