@@ -31,7 +31,7 @@ LOCAL LVAL xnotimp()
 }
 
 /* the function table */
-#ifdef _Windows    /* L. Tierney */
+#if defined(_Windows) && ! defined(WIN32)   /* L. Tierney */
 FUNDEF far funtab[] = {
 #else
 FUNDEF funtab[] = {
@@ -310,6 +310,8 @@ FUNDEF funtab[] = {
 {   "ALLOC",            S, xalloc       },
 {   "ROOM",             S, xmem         },
 {   "REGISTER-FINALIZER",S, xregfinal    },
+{   "MAKE-WEAK-BOX",    S, xmkweakbox   },
+{   "WEAK-BOX-VALUE",   S, xweakboxval  },
 #ifdef SAVERESTORE
 {   "SAVE",             S, xsave        },
 {   "RESTORE",          S, xrestore     },
@@ -856,14 +858,13 @@ FUNDEF funtab[] = {
 {   NULL,               S, xnotimp      },
 #endif /* SHAREDLIBS */
 
+{   "LOCALLY",          FM, xprogn      },
+{   "THE",              FM, xthe        },
+{   "SYMBOL-MACROLET",  FM, xsymaclet   },
+{   "XLISP::INSTALL-FUNCTION", S, xinstallfun},
+
 	/* extra table entries */
 {   NULL,               S, xnotimp      },
-{   NULL,               S, xnotimp      },
-{   NULL,               S, xnotimp      },
-{   NULL,               S, xnotimp      },
-{   NULL,               S, xnotimp      },
-{   NULL,               S, xnotimp      },
-{   NULL,               S, xnotimp      },
 {   NULL,               S, xnotimp      },
 
     /* include system dependent function pointers */
