@@ -397,15 +397,13 @@ extern FILE * _cdecl osbopen(const char *name, const char *mode);   /* open bina
 #endif
 #endif
 
-/* for the Turbo C compiler - MS-DOS, large model */
-/* Version 1.5 and 2.0.  1.5 won't compile with TIMES */
-/* Also for Turbo/Borland C++, as a C compiler */
-/* GRAPHICS ok */
-/* EDEPTH should be stacksize/24 for MEDMEM */
-#ifdef __TURBOC__
+/* MS Windows */
+#ifdef _Windows
 #ifdef BIGNUMS
-#define LDEXP myldexp
-double myldexp(double, int);
+#if defined(__TURBOC__) && defined(BIGNUMS)
+#  define LDEXP myldexp
+   double myldexp(double, int);
+#endif
 #endif
 #ifdef STSZ
 extern char *stackbase;
@@ -424,7 +422,6 @@ extern char *stackbase;
 #ifndef MSDOS
 #define MSDOS
 #endif
-#ifdef _Windows
 #include <windows.h>
 #include <windowsx.h>
 #ifdef WIN32
@@ -452,16 +449,12 @@ extern long win32stsz;
 #define MARGLO (3000)	/* Windows seems to have serious problems if stack
 			   falls below this */
 #define FNAMEMAX 127
-#endif /* _Windows */
 #ifndef WIN32
 #define MAXSLEN         (65519U)
 #define MAXVECLEN       (16383U)
 #define MAXPLEN         (32767)
 #define MAXPLEV         (32767)
 #endif /* WIN32 */
-#if __TURBOC__ < 0x297
-#define IEEEFP          /* Borland C++ V2.0 or later handles this */
-#endif
 #ifndef CDECL
 #define CDECL _Cdecl
 #endif
@@ -471,11 +464,7 @@ extern long win32stsz;
 #define OSBOPEN osbopen /* special mode for binary files */
 extern FILE * _Cdecl osbopen(const char *name, const char *mode);   /* open binary file */
 #endif
-#ifdef BIGNUMS
-#define LDEXP myldexp
-double myldexp(double, int);
-#endif
-#endif /* __TURBOC__ */
+#endif /* _Windows */
 
 /* for the JPI TopSpeed C Compiler, Medium or Large memory model */
 /* GRAPHICS ok */
