@@ -361,7 +361,8 @@ LVAL xmember(V)
 	break;
       }
 
-    if (list == slist) break;   /* list must be circular, and no match */
+    if (list == slist)  /* list must be circular, and no match */
+      xlerror("not a proper list", car (list));
   }
 
   /* restore the stack */
@@ -991,7 +992,7 @@ LVAL xnconc(V)
 }
 
 /* xsort - built-in function 'sort' */
-LOCAL LVAL mergesort P3C(LVAL, list, LVAL, sortfcn, LVAL, sortkey)
+LOCAL LVAL xlmergesort P3C(LVAL, list, LVAL, sortfcn, LVAL, sortkey)
 {
   /* Strategy: divide into two parts, (recurse) to sort each, then
      merge them together */
@@ -1018,8 +1019,8 @@ LOCAL LVAL mergesort P3C(LVAL, list, LVAL, sortfcn, LVAL, sortkey)
     rplacd(temp, NIL);		/* split left and right parts */
   }
 
-  left = mergesort(left, sortfcn, sortkey);
-  right = mergesort(right, sortfcn, sortkey);
+  left = xlmergesort(left, sortfcn, sortkey);
+  right = xlmergesort(right, sortfcn, sortkey);
 
   {
     LVAL result, resultt = NULL, leftarg, rightarg;
@@ -1094,12 +1095,12 @@ LVAL xsort()
     {
       LVAL etype = gettvecetype(list);
       list = coerce_to_list(list);
-      list = mergesort(list, sortfcn, sortkey);
+      list = xlmergesort(list, sortfcn, sortkey);
       list = coerce_to_tvec(list, etype);
       break;
     }
   case CONS:
-    list = mergesort(list, sortfcn, sortkey);
+    list = xlmergesort(list, sortfcn, sortkey);
     break;
   default: xlbadtype(list);
   }
