@@ -560,6 +560,13 @@ VOID xlprintl P3C(LVAL, fptr, LVAL, vptr, int, flag)
 	    if (flag)
 	      xlputc(fptr, '>');
 	    break;
+    case WEAKBOX:
+#ifdef PRINTCIRCLE
+	    if (checkcircle(fptr, vptr)) break;
+#endif /* PRINTCIRCLE */
+	    checkreadable(vptr);
+	    putatm(fptr,"Weak Box",vptr);
+	    break;
     case DARRAY:
 #ifdef PRINTCIRCLE
 	    if (checkcircle(fptr, vptr)) break;
@@ -1141,8 +1148,12 @@ LOCAL VOID putrndstate P3C(LVAL, fptr, LVAL, val, int, flag)
 /* putbcclosure - output a random state */
 LOCAL VOID putbcclosure P3C(LVAL, fptr, LVAL, val, int, flag)
 {
+    LVAL name;
     checkreadable(val);
-    strcpy(buf,"#<Byte-Code-Closure: #");
+    if ((name = getbcname(getbcccode(val))) != NIL)
+      sprintf(buf,"#<Byte-Code-Closure-%s: #",getstring(getpname(name)));
+    else
+      strcpy(buf,"#<Byte-Code-Closure: #");
     xlputstr(fptr,buf);
     sprintf(buf,AFMT, CVPTR(val)); xlputstr(fptr,buf);
     xlputc(fptr,'>');
