@@ -391,14 +391,8 @@ int xlread P4C(LVAL, fptr, LVAL *, pval, int, rflag, int, pwflag)
     if (! rflag)
       cleancircle(*pval, getvalue(s_rdcircdat));
 #endif /* PRINTCIRCLE */
-    /* Flush whitespace character, but better not go beyond end of current
-     * line */
-    /* I don't think this is right, so I've moved the unget to the end
-       of pname */
-    /*    if (null(getvalue(a_readpw))) {
-        int ch = xlgetc(fptr);
-        if (tentry(ch) != k_wspace) xlungetc(fptr, ch);
-    }*/
+
+    /* unbind a_readpw if necessary */
     xlunbind(olddenv);
 
     /* return status */
@@ -485,7 +479,8 @@ LVAL rmhash(V)
 		  rplaca(val,NIL);
 		  break;
 		}
-		readone(fptr,&car(val));
+		if (readone(fptr,&car(val)) == EOF)
+		  badeof(); /* Check added 3/98 */
 		rplaca(val,xleval(car(val)));
 		break;
     case 'b':
@@ -511,10 +506,11 @@ LVAL rmhash(V)
     case '\\':
 		for (i = 0; i < STRMAX-1; i++) {
 		  ch = xlgetc(fptr);  /* TAA fix to scan at end of file */
-		  if (ch == EOF || 
-		      ((tentry((unsigned char)(buf[i] = ch))  != k_const) &&
-		       (i > 0) &&      /* TAA fix for left and right paren */
-		       buf[i] != '\\' && buf[i] != '|')) {
+		  if (ch == EOF ||
+		      (buf[i] = ch,
+		       ((tentry((unsigned char)(buf[i]))  != k_const) &&
+			(i > 0) &&      /* TAA fix for left and right paren */
+			buf[i] != '\\' && buf[i] != '|'))) {
 		    xlungetc(fptr, buf[i]);
 		    break;
 		  }
@@ -604,7 +600,8 @@ LVAL rmhash(V)
     case 'C':  /* From XLISP-STAT, Copyright (c) 1988, Luke Tierney */
       {
         LVAL list;
-        readone(fptr, &list);
+        if (readone(fptr, &list) == EOF)
+	  badeof(); /* check added 3/98 */
         if (! consp(list) || ! consp(cdr(list)) || cdr(cdr(list)) != NIL)
           xlerror("bad complex number specification", list);
         rplaca(val, newcomplex(car(list), car(cdr(list))));
@@ -614,24 +611,28 @@ LVAL rmhash(V)
     case '-':  
       {
         LVAL arg;
+	int sts; /* added eof check 3/98 */
 	LVAL olddenv;
 
         xlsave1(arg);
 
 	olddenv = xldenv;
 	xldbind(s_package, xlkeypack);
-        while (! readone(fptr, &arg));
+        while (! (sts = readone(fptr, &arg)));
 	xlunbind(olddenv);
 
+        if (sts == EOF) badeof();
         if (null(getvalue(s_read_suppress)) && checkfeatures(arg, ch)) {
-          while (! readone(fptr, &arg));
-          rplaca(val, arg);
+          while (! (sts = readone(fptr, &arg)));
+          if (sts==EOF) badeof();
+	  rplaca(val, arg);
         }
         else {
 	  olddenv = xldenv;
 	  xldbind(s_read_suppress, s_true);
-          while (! readone(fptr, &arg));
-          val = NIL;
+          while (! (sts = readone(fptr, &arg)));
+          if (sts==EOF) badeof();
+	  val = NIL;
 	  xlunbind(olddenv);
         }
         xlpop();
@@ -752,6 +753,8 @@ LVAL rmhash(V)
 		  xlpop();
 		}
 		break;		  
+    case EOF:   /* added 3/98 */
+        badeof();
     default:
 /*      xlerror("illegal character after #",cvfixnum((FIXTYPE)ch)); */
 	xlerror("illegal character after #",cvchar(ch)); /*TAA Mod */
@@ -1132,6 +1135,7 @@ LOCAL LVAL pstruct P1C(LVAL, fptr)
 LOCAL LVAL pquote P2C(LVAL, fptr, LVAL, sym)
 {
     LVAL val,p;
+    int sts; /* EOF checking added 3/98 */
 
     /* protect some pointers */
     xlsave1(val);
@@ -1141,8 +1145,9 @@ LOCAL LVAL pquote P2C(LVAL, fptr, LVAL, sym)
     rplacd(val,consa(NIL));
 
     /* initialize the second to point to the quoted expression */
-    if (!xlread(fptr,&p,TRUE,FALSE))
-	badeof();
+    while ((sts = readone(fptr,&p)) == FALSE);
+    if (sts == EOF)
+        badeof();
     rplaca(cdr(val),p);
 
     /* restore the stack */
@@ -1332,7 +1337,6 @@ LOCAL int pname P2C(LVAL, fptr, int *, pescflag)
         xlfail("zero length name");     /* TAA fix, Jeff Prothero improved*/
 
     /* unget the last character and return it */
-    /* my version of whitespace-preserving */
     if (tentry(ch) != k_wspace || ! null(getvalue(a_readpw)))
       xlungetc(fptr, ch);
     return (ch);
@@ -1436,6 +1440,7 @@ LOCAL VOID badeof(V)
 }
 
 /* isnumber - check if this string is a number */
+#define DBL_DIGITS 50
 #ifdef BIGNUMS
 LOCAL int isadigit P2C(char, c, int, r)
 {
@@ -1455,6 +1460,7 @@ int isnumber P2C(char *, str, LVAL *, pval)
     int radix = 10;
     FIXTYPE numbr;		/* converted integer */
     LVAL numerp, denomp;	/* in case it is a a bignum/ratio */
+    int eoff = 0;               /* added to handle 'D' */
 
     numerp = getvalue(s_readbase);
     if (fixp(numerp)) {
@@ -1512,11 +1518,12 @@ int isnumber P2C(char *, str, LVAL *, pval)
 	}
 	/* check for an exponent */
 #ifdef READTABLECASE
-	if ((dl || dr) && (*p == 'E' || *p == 'e' || *p == 'D' || *p == 'd')) 
+	if ((dl || dr) && *p && strchr("esfdlESFDL", *p))
 #else
-	if ((dl || dr) && (*p == 'E' || *p == 'D')) 
+	if ((dl || dr) && *p && strchr("ESFDL", *p)) 
 #endif
 	    {
+	      eoff = p - str;
 	      p++;
 
 	      /* check for a sign */
@@ -1548,7 +1555,12 @@ int isnumber P2C(char *, str, LVAL *, pval)
 	  *pval = cvbratio(numerp, denomp);
         }
         else if (dr) {
-	  *pval = cvflonum(atof(str));
+	  char buf[DBL_DIGITS + 1];
+	  strncpy(buf, str, DBL_DIGITS);
+	  buf[DBL_DIGITS] = 0;
+	  if (eoff && eoff < DBL_DIGITS)
+	    buf[eoff] = 'E';
+	  *pval = cvflonum(atof(buf));
 	}
 	else {
 	  *pval = convertnumber(str, radix);
@@ -1561,6 +1573,7 @@ int isnumber P2C(char *, str, LVAL *, pval)
 {
   int dl=0, dr=0;
   char *p = str;
+  int eoff = 0;               /* added to handle 'D' */
 
   /* check for a sign */
   if (*p == '+' || *p == '-')
@@ -1579,11 +1592,12 @@ int isnumber P2C(char *, str, LVAL *, pval)
 
   /* check for an exponent */
 #ifdef READTABLECASE
-  if ((dl || dr) && (*p == 'E' || *p == 'e' || *p == 'D' || *p == 'd'))
+	if ((dl || dr) && *p && strchr("esfdlESFDL", *p))
 #else
-  if ((dl || dr) && (*p == 'E' || *p == 'D')) /* 'D' added - L. Tierney */
+	if ((dl || dr) && *p && strchr("ESFDL", *p)) 
 #endif
     {
+      eoff = p - str;
       p++;
 
       /* check for a sign */
@@ -1602,7 +1616,16 @@ int isnumber P2C(char *, str, LVAL *, pval)
   if (pval != NULL) {
     if (*str == '+') ++str;
     if (str[strlen(str)-1] == '.') str[strlen(str)-1] = 0;
-    *pval = (dr ? cvflonum(atof(str)) : cvfixnum(ICNV(str)));
+    if (dr) {
+      char buf[DBL_DIGITS + 1];
+      strncpy(buf, str, DBL_DIGITS);
+      buf[DBL_DIGITS] = 0;
+      if (eoff && eoff < DBL_DIGITS)
+	buf[eoff] = 'E';
+      *pval = cvflonum(atof(buf));
+    }
+    else
+      *pval = cvfixnum(ICNV(str));
   }
   return (TRUE);
 }
