@@ -123,6 +123,66 @@ LVAL xlgetvalue P1C(LVAL, sym)
     return (val);
 }
 
+/* xlcopytree - local version of COPY-TREE */
+LOCAL LVAL xlcopytree P1C(LVAL, x)
+{
+  if (consp(x)) {
+    LVAL val;
+#ifdef STSZ         /* This function is a good candidate for stack ov */
+    stchck();
+#endif
+    xlsave1(val);
+    val = cons(NIL, NIL);
+    rplaca(val, xlcopytree(car(x)));
+    rplacd(val, xlcopytree(cdr(x)));
+    xlpop();
+    return val;
+  }
+  else return x;
+}
+
+#define symbol_macro_binding_p(ep) \
+  cdr(car(ep)) == k_symbol_macro && consp(cdr(ep)) \
+  && car(car(cdr(ep))) == k_symbol_macro
+
+LOCAL LVAL symbol_macro_value P1C(LVAL, form)
+{
+  /* Need to copy the form in case eval splices in macros.
+     Protecting the form may not be necessary, but do it
+     anyway just to be safe. */
+  LVAL val;
+  xlprot1(form);
+  form = xlcopytree(form);
+  val = xleval(form);
+  xlpop();
+  return val;
+}
+
+LOCAL VOID set_symbol_macro P2C(LVAL, place, LVAL, val)
+{
+  /* Need to copy the place in case eval splices in macros.
+     Protecting the place may not be necessary, but do it anyway just
+     to be safe. */
+  LVAL form;
+  static LVAL s_setf = NULL;
+
+  if (s_setf == NULL) s_setf = xlenter("XLISP::SETF");
+
+  xlsave1(form);
+  xlprot1(place);
+  place = xlcopytree(place);
+
+  /* This builds up an expression of the form `(setf ,place (quote ,val)). */
+  form = consa(val);
+  form = cons(s_quote, form);
+  form = consa(form);
+  form = cons(place, form);
+  form = cons(s_setf, form);
+
+  xleval(form);
+  xlpopn(2);
+}
+
 /* xlxgetvalue - get the value of a symbol */
 LVAL xlxgetvalue P1C(LVAL, sym)
 {
@@ -141,12 +201,18 @@ LVAL xlxgetvalue P1C(LVAL, sym)
 	/* check an environment stack frame */
 	else {
 	    for (; consp(ep); ep = cdr(ep))
-		if (sym == car(car(ep)))
-		    return (cdr(car(ep)));
+	      if (sym == car(car(ep))) {
+		if (symbol_macro_binding_p(ep))
+		  return symbol_macro_value(cdr(car(cdr(ep))));
+		else return cdr(car(ep));
+	      }
 	}
 
     /* return the global value */
-    return (getvalue(sym));
+    if (! boundp(sym) && ! specialp(sym) &&
+	(ep = findprop(getplist(sym), k_symbol_macro)) != NIL)
+      return symbol_macro_value(car(ep));
+    else return (getvalue(sym));
 }
 
 /* xlsetvalue - set the value of a symbol */
@@ -172,13 +238,20 @@ VOID xlsetvalue P2C(LVAL, sym, LVAL, val)
 	else {
 	    for (; consp(ep); ep = cdr(ep))
 		if (sym == car(car(ep))) {
-		    rplacd(car(ep),val);
+		    if (symbol_macro_binding_p(ep))
+		      set_symbol_macro(cdr(car(cdr(ep))), val);
+		    else
+		      rplacd(car(ep),val);
 		    return;
 		}
 	}
 
     /* store the global value */
-    setvalue(sym,val);
+    if (! boundp(sym) && ! specialp(sym) &&
+	(ep = findprop(getplist(sym), k_symbol_macro)) != NIL)
+      set_symbol_macro(car(ep), val);
+    else
+      setvalue(sym,val);
 }
 
 /* xlgetfunction - get the functional value of a symbol (with check) */
@@ -288,7 +361,7 @@ int xlhash P2C(LVAL, obj, int, len)
             swizzle.i = getfixnum(obj);
             return (int) (swizzle.k % len);
         case FLONUM:
-            swizzle.j = getflonum(obj);
+            swizzle.j = (float)(getflonum(obj));
             return (int) (swizzle.k % len);
         case CHAR:
             return getchcode(obj) % len;
@@ -1311,3 +1384,27 @@ LVAL xsympackage(V)
   return (getpackage(sym));
 }
 #endif /* PACKAGES */
+
+LVAL xinstallfun(V)
+{
+  LVAL symbol = xlgasymbol();
+  LVAL fun = xlgetarg();
+  LVAL def = moreargs() ? xlgetarg() : NIL;
+  LVAL env = moreargs() ? xlgetarg() : NIL;
+
+  xllastarg();
+  switch (ntype(fun)) {
+  case BCCLOSURE:
+    if (getbcname(getbcccode(fun)) == NIL)
+      setbcname(getbcccode(fun), symbol);
+    if (def != NIL && getbcdef(getbcccode(fun)) == NIL)
+      setbcdef(getbcccode(fun), cons(def,env));
+    break;
+  case CLOSURE:
+    if (getname(fun) == NIL)
+      setname(fun, symbol);
+    break;
+  }
+  setfunction(symbol, fun);
+  return symbol;
+}
